# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# *********************************************  GatingProjection ******************************************************

"""
Contents
--------

  * `GatingProjection_Overview`
  * `GatingProjection_Creation`
      - `GatingProjection_Deferred_Initialization`
  * `GatingProjection_Structure`
  * `GatingProjection_Execution`
  * `GatingProjection_Class_Reference`


.. _GatingProjection_Overview:

Overview
--------

A GatingProjection is a type of `ModulatoryProjection <ModulatoryProjection>` that projects to the `InputPort` or
`OutputPort` of a `Mechanism <Mechanism>`. It takes the value of a `GatingSignal` of a `GatingMechanism`,
and uses it to modulate the `value <Port_Base.value>` of the Port to which it projects.

.. _GatingProjection_Creation:

Creating a GatingProjection
----------------------------

A GatingProjection can be created using any of the standard ways to `create a projection <Projection_Creation>`,
or by including it in the specification of an `InputPort <InputPort_Projection_Source_Specification>` or `OutputPort
<OutputPort_Projections>` .  If a GatingProjection is created explicitly (using its constructor), its **receiver**
argument can be specified as a particular InputPort or OutputPort of a designated `Mechanism <Mechanism>`, or simply
as the Mechanism.  In the latter case, the Mechanism's `primary InputPort <InputPort_Primary>` will be used. If the
GatingProjection is included in an InputPort or OutputPort specification, that Port will be assigned as the
GatingProjection's `receiver <GatingProjection.receiver>`. If the **sender** and/or **receiver** arguments are not
specified, its initialization is `deferred <GatingProjection_Deferred_Initialization>`.


.. _GatingProjection_Deferred_Initialization:

*Deferred Initialization*
~~~~~~~~~~~~~~~~~~~~~~~~~

When a GatingProjection is created, its full initialization is `deferred <Component_Deferred_Init>` until its
`sender <ControlProjection.sender>` and `receiver <ControlProjection.receiver>` have been fully specified.  This allows
a GatingProjection to be created before its `sender` and/or `receiver` have been created (e.g., before them in a
script), by calling its constructor without specifying its **sender** or **receiver** arguments. However, for the
GatingProjection to be operational, initialization must be completed by calling its `deferred_init` method.  This is
not necessary if the Port(s) to be gated are specified in the **gate** argument of a `GatingMechanism
<GatingMechanism_Specifying_Gating>`, in which case deferred initialization is completed automatically by the
GatingMechanism when it is created.

.. _GatingProjection_Structure:

Structure
---------

The `sender <GatingProjection.sender>` of a GatingProjection is a `GatingSignal` of a `GatingMechanism`.  The `value
<GatingSignal.value>` of the `sender <GatingProjection.sender>` is used by the GatingProjection as its
`variable <GatingProjection.variable>`;  this is also assigned to its `gating_signal
<GatingProjection.gating_signal>` attribute, and serves as the input to the GatingProjection's `function
<Projection_Base.function>`.  The default `function <Projection_Base.function>` for a
GatingProjection is an identity function (`Linear` with **slope**\\ =1 and **intercept**\\ =0);  that is,
it simply conveys the value of its `gating_signal <GatingProjection.gating_signal>` to its `receiver
<GatingProjection.receiver>`, for use in modifying the `value <Port_Base.value>` of the Port that it gates. Its
`receiver <GatingProjection.receiver>` is the `InputPort` or `OutputPort` of a `Mechanism <Mechanism>`.

.. _GatingProjection_Execution:

Execution
---------

A GatingProjection cannot be executed directly.  It is executed when the `InputPort` or `OutputPort` to which it
projects is updated.  Note that this only occurs when the `Mechanism <Mechanism>` to which the `Port <Port>` belongs
is executed (see `Lazy Evaluation <Component_Lazy_Updating>` for an explanation of "lazy" updating). When a
GatingProjection is executed, its `function <Projection_Base.function>` gets the `gating_signal
<GatingProjection.gating_signal>` from its `sender <GatingProjection.sender>` and conveys that to its `receiver
<GatingProjection.receiver>`.  This is used by the `receiver <GatingProjection.receiver>` to modify the `value
<Port_Base.value>` of the Port gated by the GatingProjection (see `ModulatorySignal_Modulation`, `InputPort Execution
<InputPort_Execution>` and `OutputPort Execution <OutputPort_Execution>` for how modulation operates and how this
applies to a InputPorts and OutputPorts).

.. note::
   The changes in an InputPort or OutputPort's `value <Port_Base.value >` in response to the execution of a
   GatingProjection are not applied until the Mechanism to which the Port belongs is next executed;
   see `Lazy Evaluation <Component_Lazy_Updating>` for an explanation of "lazy" updating).

.. _GatingProjection_Class_Reference:

Class Reference
---------------

"""
from beartype import beartype

from psyneulink._typing import Optional, Mapping

from psyneulink.core.components.component import parameter_keywords
from psyneulink.core.components.functions.function import FunctionOutputType
from psyneulink.core.components.functions.nonstateful.transferfunctions import Linear
from psyneulink.core.components.mechanisms.modulatory.control.gating.gatingmechanism import GatingMechanism
from psyneulink.core.components.projections.modulatory.modulatoryprojection import ModulatoryProjection_Base
from psyneulink.core.components.projections.projection import ProjectionError, projection_keywords
from psyneulink.core.components.shellclasses import Mechanism, Process_Base
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.keywords import \
    FUNCTION_OUTPUT_TYPE, GATE, GATING_MECHANISM, GATING_PROJECTION, GATING_SIGNAL, \
    INPUT_PORT, OUTPUT_PORT
from psyneulink.core.globals.parameters import Parameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel

__all__ = [
    'GATING_SIGNAL_PARAMS', 'GatingProjection', 'GatingProjectionError',
]

parameter_keywords.update({GATING_PROJECTION, GATE})
projection_keywords.update({GATING_PROJECTION, GATE})
GATING_SIGNAL_PARAMS = 'gating_signal_params'

class GatingProjectionError(ProjectionError):
    pass


def _gating_signal_getter(owning_component=None, context=None):
    return owning_component.sender.parameters.value._get(context)


def _gating_signal_setter(value, owning_component=None, context=None):
    owning_component.sender.parameters.value._set(value, context)
    return value


class GatingProjection(ModulatoryProjection_Base):
    """
    GatingProjection(           \
     sender=None,               \
     receiver=None,             \
     gating_signal_params=None)

    Subclass of `ModulatoryProjection <ModulatoryProjection>` that modulates the value of an `InputPort` or
    `OutputPort` of a `Mechanism <Mechanism>`.
    See `Projection <ModulatoryProjection_Class_Reference>` for additional arguments and attributes.


    Arguments
    ---------

    sender : GatingMechanism or GatingSignal : default None
        specifies the source of the `gating_signal <GatingProjection.gating_signal>` for the GatingProjection;
        if it is not specified and cannot be `inferred from context <GatingProjection_Creation>` , initialization is
        `deferred <GatingProjection_Deferred_Initialization>`.

    receiver : Mechanism, InputPort or OutputPort : default None
        specifies the `InputPort` or `OutputPort` to which the GatingProjection projects; if it is not specified,
        and cannot be `inferred from context <GatingProjection_Creation>`, initialization is `deferred
        <GatingProjection_Deferred_Initialization>`.

    gating_signal_params : Dict[param keyword: param value]
        a `parameter dictionary <ParameterPort_Specification>` that can be used to specify the parameters for the
        GatingProjection's `sender <GatingProjection.sender>` (see `GatingSignal_Structure` for a description
        of GatingSignal parameters).

    Attributes
    ----------

    sender : GatingSignal
        source of the `gating_signal <GatingProjection.gating_signal>`.

    receiver : InputPort or OutputPort of a Mechanism
        `InputPort` or `OutputPort` to which the GatingProjection projects.

    variable : 2d np.array
        same as `gating_signal <GatingProjection.gating_signal>`.

    gating_signal : 1d np.array
        the `value <GatingSignal.value>` of the GatingProjection's `sender <GatingProjection.sender>`.

    value : float
        the value used to modify the `value <Port_Base.value>` of the `InputPort` or `OutputPort` gated by the
        GatingProjection (see `ModulatorySignal_Modulation`, `InputPort Execution <InputPort_Execution>`, and
        `OutputPort Execution <OutputPort_Execution>` for how modulation operates and how this applies to InputPorts
        and OutputPorts).

    """

    color = 0

    componentType = GATING_PROJECTION
    className = componentType
    suffix = " " + className

    classPreferenceLevel = PreferenceLevel.TYPE

    class sockets:
        sender=[GATING_SIGNAL]
        receiver=[INPUT_PORT, OUTPUT_PORT]

    class Parameters(ModulatoryProjection_Base.Parameters):
        """
            Attributes
            ----------

                function
                    see `function <GatingProjection.function>`

                    :default value: `Linear`(output_type=0)
                    :type: `Function`

                gating_signal
                    see `gating_signal <GatingProjection.gating_signal>`

                    :default value: None
                    :type:
                    :read only: True
        """
        function = Parameter(Linear(params={FUNCTION_OUTPUT_TYPE: FunctionOutputType.NP_0D_ARRAY}), stateful=False, loggable=False)
        gating_signal = Parameter(None, read_only=True, getter=_gating_signal_getter, setter=_gating_signal_setter, pnl_internal=True)

        gating_signal_params = Parameter(
            None,
            stateful=False,
            loggable=False,
            read_only=True,
            user=False,
            pnl_internal=True
        )

    projection_sender = GatingMechanism

    @check_user_specified
    @beartype
    def __init__(self,
                 sender=None,
                 receiver=None,
                 function=None,
                 weight=None,
                 exponent=None,
                 gating_signal_params:Optional[Mapping]=None,
                 params=None,
                 name=None,
                 prefs:   Optional[ValidPrefSet] = None,
                 **kwargs
                 ):
        # If receiver has not been assigned, defer init to Port.instantiate_projection_to_state()
        if sender is None or receiver is None:
            # Flag for deferred initialization
            self.initialization_status = ContextFlags.DEFERRED_INIT

        # Validate sender (as variable) and params, and assign to variable
        # Note: pass name of mechanism (to override assignment of componentName in super.__init__)
        super().__init__(
            sender=sender,
            receiver=receiver,
            weight=weight,
            exponent=exponent,
            function=function,
            gating_signal_params=gating_signal_params,
            params=params,
            name=name,
            prefs=prefs,
            **kwargs
        )

    def _instantiate_sender(self, sender, params=None, context=None):
        """Check that sender is not a process and that, if specified as a Mechanism, it is a GatingMechanism
        """

        # A Process can't be the sender of a GatingProjection
        if isinstance(sender, Process_Base):
            raise ProjectionError(
                "PROGRAM ERROR: attempt to add a {} from a Process {0} "
                "to a Mechanism {0} in pathway list".format(
                    GATING_PROJECTION, self.name, sender.name
                )
            )

        # If sender is specified as a Mechanism, validate that it is a GatingMechanism
        if isinstance(sender, Mechanism):
            if not isinstance(sender, GatingMechanism):
                raise GatingProjectionError(
                    "Mechanism specified as sender for {} ({}) must be a {} (but it is a {})".format(
                        GATING_MECHANISM, self.name, sender.name, sender.__class__.__name__
                    )
                )

        # Call super to instantiate sender
        super()._instantiate_sender(sender, context=context)

    def _validate_params(self, request_set, target_set=None, context=None):

        super()._validate_params(request_set=request_set, target_set=target_set, context=context)

        if self.initialization_status == ContextFlags.INITIALIZING:
            from psyneulink.core.components.ports.inputport import InputPort
            from psyneulink.core.components.ports.outputport import OutputPort
            if not isinstance(self.receiver, (InputPort, OutputPort, Mechanism)):
                raise GatingProjectionError("Receiver specified for {} {} is not a "
                                            "Mechanism, InputPort or OutputPort".
                                            format(self.receiver, self.name))

    def _instantiate_receiver(self, context=None):
        """Assign port if receiver is Mechanism, and match output to param being modulated
        """
        # If receiver specification was a Mechanism, re-assign to the mechanism's primary inputPort
        if isinstance(self.receiver, Mechanism):
            # If Mechanism is specified as receiver, assign GatingProjection to primary inputPort as the default
            self.receiver = self.receiver.input_ports[0]

        # # Match type of GatingProjection.value to type to the parameter being modulated
        # modulated_param = self.sender.modulation
        # function = self.receiver.function
        # function_param = function.modulated_param
        # function_param_value = function.function_param
        # gating_projection_function = self.function
        # gating_projection_function.output_type = type(function_param_value)
        # # ASSIGN FUNCTION TYPE TO FUNCTION HERE

        super()._instantiate_receiver(context=context)

    @property
    def gating_signal(self):
        return self.sender.value



import numpy as np
import pytest

import psyneulink as pnl
from psyneulink.core.components.functions.nonstateful.objectivefunctions import Distance
from psyneulink.core.components.functions.nonstateful.optimizationfunctions import GridSearch, MINIMIZE
from psyneulink.core.components.functions.nonstateful.transferfunctions import GaussianDistort
from psyneulink.core.components.mechanisms.modulatory.control import OptimizationControlMechanism
from psyneulink.core.components.mechanisms.processing.objectivemechanism import ObjectiveMechanism
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism
from psyneulink.core.components.mechanisms.processing.transfermechanism import TransferMechanism
#from psyneulink.core.components.ports.inputport import SHADOW_INPUTS
from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
from psyneulink.core.compositions.composition import Composition, NodeRole
from psyneulink.core.globals.keywords import VARIANCE, NORMED_L0_SIMILARITY
from psyneulink.core.globals.utilities import _SeededPhilox
from psyneulink.library.components.mechanisms.processing.objective.comparatormechanism import ComparatorMechanism


@pytest.mark.model
@pytest.mark.benchmark(group="Greedy Agent")
def test_simplified_greedy_agent(benchmark, comp_mode):
    # These should probably be replaced by reference to ForagerEnv constants:
    obs_len = 2
    action_len = 2
    player_coord_idx = slice(0,2)
    predator_coord_idx = slice(3,5)
    prey_coord_idx = slice(6,8)
    player_value_idx = 2
    predator_value_idx = 5
    prey_value_idx = 8

    player_len = prey_len = predator_len = obs_len

# The original needs GaussianDistort
#    player = ProcessingMechanism(input_shapes=prey_len, function=GaussianDistort, name="PLAYER OBS")
#    prey = ProcessingMechanism(input_shapes=prey_len, function=GaussianDistort, name="PREY OBS")
    player = TransferMechanism(input_shapes=prey_len, name="PLAYER OBS")
    prey = TransferMechanism(input_shapes=prey_len, name="PREY OBS")

    # Use ComparatorMechanism to compute direction of action as difference of coordinates between player and prey:
    # note: unitization is done in main loop, to allow compilation of LinearCombination function) (TBI)
    greedy_action_mech = ComparatorMechanism(name='MOTOR OUTPUT',sample=player,target=prey)

    agent_comp = Composition(name='PREDATOR-PREY COMPOSITION')
    agent_comp.add_node(player)
    agent_comp.add_node(prey)
    agent_comp.add_node(greedy_action_mech)

    # Projections to greedy_action_mech were created by assignments of sample and target args in its constructor,
    #  so just add them to the Composition).
    for projection in greedy_action_mech.projections:
        agent_comp.add_projection(projection)

    run_results = benchmark(agent_comp.run, inputs={player:[[619,177]],prey:[[419,69]]}, execution_mode=comp_mode)
    np.testing.assert_allclose(run_results, [[-200, -108]])

@pytest.mark.model
@pytest.mark.benchmark(group="Greedy Agant Random")
def test_simplified_greedy_agent_random(benchmark, comp_mode):
    # These should probably be replaced by reference to ForagerEnv constants:
    obs_len = 2
    action_len = 2
    player_coord_idx = slice(0,2)
    predator_coord_idx = slice(3,5)
    prey_coord_idx = slice(6,8)
    player_value_idx = 2
    predator_value_idx = 5
    prey_value_idx = 8

    player_len = prey_len = predator_len = obs_len

    player = ProcessingMechanism(input_shapes=prey_len, function=GaussianDistort, name="PLAYER OBS")
    prey = ProcessingMechanism(input_shapes=prey_len, function=GaussianDistort, name="PREY OBS")

    # Use ComparatorMechanism to compute direction of action as difference of coordinates between player and prey:
    # note: unitization is done in main loop, to allow compilation of LinearCombination function) (TBI)
    greedy_action_mech = ComparatorMechanism(name='MOTOR OUTPUT',sample=player,target=prey)

    agent_comp = Composition(name='PREDATOR-PREY COMPOSITION')
    agent_comp.add_node(player)
    agent_comp.add_node(prey)
    agent_comp.add_node(greedy_action_mech)

    # Projections to greedy_action_mech were created by assignments of sample and target args in its constructor,
    #  so just add them to the Composition).
    for projection in greedy_action_mech.projections:
        agent_comp.add_projection(projection)

    run_results = benchmark(agent_comp.run, inputs={player:[[619, 177]], prey:[[419, 69]]}, execution_mode=comp_mode)
    np.testing.assert_allclose(run_results, [[-199.5484223217141, -107.79361870517444]])

@pytest.mark.model
@pytest.mark.benchmark(group="Predator Prey")
@pytest.mark.parametrize("mode, ocm_mode", pytest.helpers.get_comp_and_ocm_execution_modes())
@pytest.mark.parametrize("samples", [[0,10],
    pytest.param([0,3,6,10], marks=pytest.mark.stress),
    pytest.param([0,2,4,6,8,10], marks=pytest.mark.stress),
    pytest.param([a / 10.0 for a in range(0, 101)], marks=pytest.mark.stress),
], ids=lambda x: len(x))
@pytest.mark.parametrize('prng', ['Default', 'Philox'])
@pytest.mark.parametrize('fp_type', [pnl.core.llvm.ir.DoubleType, pnl.core.llvm.ir.FloatType])
def test_predator_prey(benchmark, mode, ocm_mode, prng, samples, fp_type):
    if len(samples) > 10 and mode not in {pnl.ExecutionMode.LLVM,
                                          pnl.ExecutionMode.LLVMExec,
                                          pnl.ExecutionMode.LLVMRun} and \
       ocm_mode not in {'LLVM', 'PTX'}:
        pytest.skip("This test takes too long")

    # Instantiate LLVMBuilderContext using the preferred fp type
    pnl.core.llvm.builder_context.LLVMBuilderContext(fp_type())

    benchmark.group = "Predator-Prey " + str(len(samples))
    obs_len = 3
    obs_coords = 2

    player_len = prey_len = predator_len = obs_coords

    # Input Mechanisms
    player_pos = ProcessingMechanism(input_shapes=player_len, name="PLAYER POS")
    prey_pos = ProcessingMechanism(input_shapes=prey_len, name="PREY POS")
    predator_pos = ProcessingMechanism(input_shapes=predator_len, name="PREDATOR POS")

    # Perceptual Mechanisms
    player_obs = ProcessingMechanism(input_shapes=prey_len, function=GaussianDistort, name="PLAYER OBS")
    prey_obs = ProcessingMechanism(input_shapes=prey_len, function=GaussianDistort, name="PREY OBS")
    predator_obs = TransferMechanism(input_shapes=predator_len, function=GaussianDistort, name="PREDATOR OBS")


    def action_fn(variable):
        predator_pos = variable[0]
        player_pos = variable[1]
        prey_pos = variable[2]

        # Directions away from predator and towards prey
        pred_2_player = player_pos - predator_pos
        play_2_prey = prey_pos - player_pos

        # Distances to predator and prey
        distance_predator = np.sqrt(pred_2_player[0] * pred_2_player[0] + pred_2_player[1] * pred_2_player[1])
        distance_prey = np.sqrt(play_2_prey[0] * play_2_prey[0] + play_2_prey[1] * play_2_prey[1])

        # Normalized directions from predator and towards prey
        pred_2_player_norm = pred_2_player / distance_predator
        play_2_prey_norm = play_2_prey / distance_prey

        # Weighted directions from predator and towards prey
        # weights are reversed so closer agent has greater impact on movement
        pred_2_player_n = pred_2_player_norm * (distance_prey / (distance_predator + distance_prey))
        play_2_prey_n = play_2_prey_norm * (distance_predator / (distance_predator + distance_prey))

        return pred_2_player_n + play_2_prey_n

    # note: unitization is done in main loop
    greedy_action_mech = pnl.ProcessingMechanism(function=action_fn, input_ports=["predator", "player", "prey"],
                                                 default_variable=[[0, 1], [0, -1], [1, 0]], name="ACTION")

    direct_move = ComparatorMechanism(name='DIRECT MOVE',sample=player_pos, target=prey_pos)

    # Create Composition
    agent_comp = Composition(name='PREDATOR-PREY COMPOSITION')
    agent_comp.add_linear_processing_pathway([player_pos, player_obs])
    agent_comp.add_linear_processing_pathway([prey_pos, prey_obs])
    agent_comp.add_linear_processing_pathway([predator_pos, predator_obs])
    agent_comp.add_node(greedy_action_mech)
    agent_comp.add_node(direct_move)
    agent_comp.add_projection(pnl.MappingProjection(predator_obs, greedy_action_mech.input_ports[0]))
    agent_comp.add_projection(pnl.MappingProjection(prey_obs, greedy_action_mech.input_ports[1]))
    agent_comp.add_projection(pnl.MappingProjection(player_obs, greedy_action_mech.input_ports[2]))
    agent_comp.exclude_node_roles(direct_move, NodeRole.OUTPUT)


    ocm = OptimizationControlMechanism(state_features=[player_pos, prey_pos, predator_pos],
    # ocm = OptimizationControlMechanism(state_features={SHADOW_INPUTS: [player_pos, prey_pos, predator_pos]},
                                       agent_rep=agent_comp,
                                       function=GridSearch(direction=MINIMIZE,
                                                           save_values=True),

                                       objective_mechanism=ObjectiveMechanism(function=Distance(metric=NORMED_L0_SIMILARITY),
                                                                              monitor=[
                                                                                  greedy_action_mech,
                                                                                  direct_move
                                                                              ]),
                                       control_signals=[ControlSignal(modulates=(VARIANCE, player_obs),
                                                                      allocation_samples=samples,
                                                                      cost_options=pnl.CostFunctions.INTENSITY),
                                                        ControlSignal(modulates=(VARIANCE, predator_obs),
                                                                      allocation_samples=samples,
                                                                      cost_options=pnl.CostFunctions.INTENSITY),
                                                        ControlSignal(modulates=(VARIANCE, prey_obs),
                                                                      allocation_samples=samples,
                                                                      cost_options=pnl.CostFunctions.INTENSITY),
                                                        ],
                                       )
    agent_comp.add_controller(ocm)
    agent_comp.enable_controller = True
    ocm.comp_execution_mode = ocm_mode

    if prng == 'Philox':
        player_obs.function.parameters.random_state.set(_SeededPhilox([0]))
        prey_obs.function.parameters.random_state.set(_SeededPhilox([0]))
        predator_obs.function.parameters.random_state.set(_SeededPhilox([0]))
        ocm.function.parameters.random_state.set(_SeededPhilox([0]))

    input_dict = {player_pos:[[1.1576537,  0.60782117]],
                  predator_pos:[[-0.03479106, -0.47666293]],
                  prey_pos:[[-0.60836214,  0.1760381 ]],
                 }
    run_results = benchmark(agent_comp.run, inputs=input_dict, num_trials=2, execution_mode=mode)

    if len(samples) == 2:
        # FIX: REQUIRES TOLERANCE OF 1e-6
        if prng == 'Default':
            # np.testing.assert_allclose(run_results[0], [[0.9705216285127504, -0.1343332460369043]])
            # np.testing.assert_allclose(run_results, [[0.9705216285127504, -0.1343332460369043]])
            np.testing.assert_allclose(run_results, [[0.9705216285127504, -0.1343332460369043]], atol=1e-6, rtol=1e-6)
        elif prng == 'Philox':
            if mode == pnl.ExecutionMode.Python or pytest.helpers.llvm_current_fp_precision() == 'fp64':
                # np.testing.assert_allclose(run_results[0], [[-0.16882940384606543, -0.07280074899749223]])
                np.testing.assert_allclose(run_results, [[-0.16882940384606543, -0.07280074899749223]])
            elif pytest.helpers.llvm_current_fp_precision() == 'fp32':
                # np.testing.assert_allclose(run_results[0], [[-0.8639436960220337, 0.4983368515968323]])
                np.testing.assert_allclose(run_results, [[-0.8639436960220337, 0.4983368515968323]])
            else:
                assert False, "Unkown FP type!"
        else:
            assert False, "Unknown PRNG!"

        if mode == pnl.ExecutionMode.Python and not benchmark.enabled:
            # FIXME: The results are 'close' for both Philox and MT,
            #        because they're dominated by costs
            # FIX: Requires 1e-5 tolerance
            np.testing.assert_allclose(np.asarray(ocm.function.saved_values).flatten(),
                                       [-2.66258741, -22027.9970321, -22028.17515945, -44053.59867802,
                                        -22028.06045185, -44053.4048842, -44053.40736234, -66078.90687915],
                                       rtol=1e-5, atol=1e-5)



"""
DECLAN Params: **************************************************************************
√ episodic_lr = 1  # learning rate for the episodic pathway
√ temperature = 0.1  # temperature for EM retrieval (lower is more argmax-like)
√ n_optimization_steps = 10  # number of update steps
sim_thresh = 0.8 # threshold for discarding bad seeds -- can probably ignore this for now
Filter runs whose context representations are too uniform (i.e. not similar to "checkerboard" foil)

May need to pad the context reps because there will be 999 reps
def filter_run(run_em, thresh=0.8):
    foil = np.zeros([4,4])
    foil[::2, ::2] = 1
    foil[1::2, 1::2] = 1
    run_em = run_em.reshape(200, 5, 11).mean(axis=1)
    mat = cosine_similarity(run_em, run_em)
    vec = mat[:160, :160].reshape(4, 40, 4, 40).mean(axis=(1, 3)).ravel()
    return cosine_similarity(foil.reshape(1, -1), vec.reshape(1, -1))[0][0]

# Stack the model predictions (should be 999x11), pad with zeros, and reshape into trials for averaging.
em_preds = np.vstack([em_preds, np.zeros([1,11])]).reshape(-1,5,11)

# Stack the ground truth states (should be 999x11), pad with zeros, and reshape into trials for averaging.
ys = np.vstack([data_loader.dataset.ys.cpu().numpy(), np.zeros([1,11])]).reshape(-1,5,11)

# compute the probability as a performance metric
def calc_prob(em_preds, test_ys):
    em_preds, test_ys = em_preds[:, 2:-1, :], test_ys[:, 2:-1, :]
    em_probability = (em_preds*test_ys).sum(-1).mean(-1)
    trial_probs = (em_preds*test_ys)
    return em_probability, trial_probs

Calculate the retrieval probability of the correct response as a performance metric (probs)
probs, trial_probs = calc_prob(em_preds, test_ys)
"""
from psyneulink.core.llvm import ExecutionMode
from psyneulink.core.globals.keywords import ALL, ADAPTIVE, CONTROL, CPU, Loss, MPS, OPTIMIZATION_STEP, RUN, TRIAL

model_params = dict(

    # Names:
    name = "EGO Model CSW",
    em_name = "EM",
    state_input_layer_name = "STATE",
    previous_state_layer_name = "PREVIOUS STATE",
    context_layer_name = 'CONTEXT',
    prediction_layer_name = "PREDICTION",

    # Structural
    state_d = 11, # length of state vector
    previous_state_d = 11, # length of state vector
    context_d = 11, # length of context vector
    memory_capacity = ALL, # number of entries in EM memory; ALL=> match to number of stims
    memory_init = (0,.0001),  # Initialize memory with random values in interval
    # memory_init = None,  # Initialize with zeros
    # concatenate_queries = False,
    concatenate_queries = True,

    # environment
    # curriculum_type = 'Interleaved',
    curriculum_type = 'Blocked',
    # num_stims = 100,  # Integer or ALL
    num_stims = ALL,  # Integer or ALL

    # Processing
    integration_rate = .69, # rate at which state is integrated into new context
    # state_weight =normalize_field_weightsnormalize_field_weights 1, # weight of the state used during memory retrieval
    # context_weight = 1, # weight of the context used during memory retrieval
    previous_state_weight = .5, # weight of the state used during memory retrieval
    context_weight = .5, # weight of the context used during memory retrieval
    state_weight = None, # weight of the state used during memory retrieval
    # normalize_field_weights = False, # whether to normalize the field weights during memory retrieval
    normalize_field_weights = True, # whether to normalize the field weights during memory retrieval
    normalize_memories = False, # whether to normalize the memory during memory retrieval
    # normalize_memories = True, # whether to normalize the memory during memory retrieval
    normalize_memories = False, # whether to normalize the memory vectors
    # softmax_temperature = None, # temperature of the softmax used during memory retrieval (smaller means more argmax-like
    softmax_temperature = .1, # temperature of the softmax used during memory retrieval (smaller means more argmax-like
    # softmax_temperature = ADAPTIVE, # temperature of the softmax used during memory retrieval (smaller means more argmax-like
    # softmax_temperature = CONTROL, # temperature of the softmax used during memory retrieval (smaller means more argmax-like
    # softmax_threshold = None, # threshold used to mask out small values in softmax
    softmax_threshold = .001, # threshold used to mask out small values in softmax
    # target_fields=[True, False, False], # Enable learning for PREDICTION (STATE) but not CONTEXT or PREVIOUS STATE
    enable_learning = True,
    loss_spec = Loss.BINARY_CROSS_ENTROPY,
    # loss_spec = Loss.MSE,
    learning_rate = .5,
    # num_optimization_steps = 1,
    num_optimization_steps = 10,
    synch_weights = RUN,
    synch_values = RUN,
    synch_results = RUN,
    # execution_mode = ExecutionMode.Python,
    execution_mode = ExecutionMode.PyTorch,
    device = CPU,
    # device = MPS,
)
#endregion


import numpy as np
import pytest

import psyneulink as pnl

from psyneulink.core.components.functions.nonstateful.transferfunctions import Identity, Linear
from psyneulink.core.components.mechanisms.processing.compositioninterfacemechanism import CompositionInterfaceMechanism
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism
from psyneulink.core.components.mechanisms.processing.transfermechanism import TransferMechanism
from psyneulink.core.components.mechanisms.modulatory.control.controlmechanism import ControlMechanism
from psyneulink.core.components.mechanisms.modulatory.control.optimizationcontrolmechanism import OptimizationControlMechanism
from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
from psyneulink.core.components.ports.inputport import InputPort
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection
from psyneulink.core.compositions.composition import Composition, CompositionError, NodeRole
from psyneulink.core.scheduling.scheduler import Scheduler
from psyneulink.core.globals.utilities import convert_all_elements_to_np_array
from psyneulink.core.globals.keywords import INTERCEPT, NOISE, SLOPE


class TestExecuteCIM:

    def test_identity_function(self):

        I = Identity()

        output = I.execute(2.0)
        assert output == 2.0

        output = I.execute([1.0, 2.0,3.0])
        np.testing.assert_allclose([1.0, 2.0,3.0], output)

        output = I.execute([[1.0, 2.0], [3.0]])

        np.testing.assert_allclose([1.0, 2.0], output[0])
        np.testing.assert_allclose([3.0], output[1])

    def test_standalone_CIM(self):

        cim = CompositionInterfaceMechanism()
        cim.execute(2.0)
        np.testing.assert_allclose(cim.value, [[2.0]])

    def test_assign_value(self):
        cim = CompositionInterfaceMechanism()
        cim.defaults.variable = [2.0]
        cim.execute()
        np.testing.assert_allclose(cim.value, [[2.0]])

    def test_standalone_CIM_multiple_input_ports(self):

        cim = CompositionInterfaceMechanism(default_variable=[[0.0], [0.0], [0.0]])
        cim.execute([[1.0], [2.0], [3.0]])
        np.testing.assert_allclose(cim.value, [[1.0], [2.0], [3.0]])

    def test_standalone_processing_multiple_input_ports(self):

        processing_mech = ProcessingMechanism(default_variable=[[0.0], [0.0], [0.0]])
        processing_mech.execute([[1.0], [2.0], [3.0]])
        np.testing.assert_allclose(processing_mech.value, [[1.0], [2.0], [3.0]])


    def test_one_input_port_one_output_port(self):

        comp = Composition()

        A = TransferMechanism(name="A",
                              function=Linear(slope=2.0))

        B = TransferMechanism(name="B",
                              function=Linear(slope=3.0))

        comp.add_node(A)
        comp.add_node(B)

        comp.add_projection(MappingProjection(sender=A, receiver=B), A, B)

        inputs_dict = {
            A: [[5.]],
        }
        sched = Scheduler(composition=comp)
        output = comp.run(
            inputs=inputs_dict,
            scheduler=sched
        )

        np.testing.assert_allclose([[30]], output)

    def test_two_input_ports_two_output_ports(self):

        comp = Composition()

        A = TransferMechanism(name="A",
                              default_variable=[[0.0], [0.0]],
                              function=Linear(slope=2.0))

        B = TransferMechanism(name="B",
                              default_variable=[[0.0], [0.0]],
                              function=Linear(slope=3.0))

        comp.add_node(A)
        comp.add_node(B)

        comp.add_projection(MappingProjection(sender=A, receiver=B), A, B)
        comp.add_projection(MappingProjection(sender=A.output_ports[1], receiver=B.input_ports[1]), A, B)

        inputs_dict = {
            A: [[5.], [6.]],
        }
        sched = Scheduler(composition=comp)
        output = comp.run(
            inputs=inputs_dict,
            scheduler=sched
        )

        np.testing.assert_allclose([[30.], [36.]], output)


        # np.testing.assert_allclose([30.], comp.output_CIM.output_ports[1].value)
        # np.testing.assert_allclose([36.], comp.output_CIM.output_ports[2].value)

class TestConnectCompositionsViaCIMS:

    @pytest.mark.nested
    @pytest.mark.composition
    def test_connect_compositions_with_simple_states(self, comp_mode):

        comp1 = Composition(name="first_composition")

        A = TransferMechanism(name="A",
                              function=Linear(slope=2.0))

        B = TransferMechanism(name="B",
                              function=Linear(slope=3.0))

        comp1.add_node(A)
        comp1.add_node(B)

        comp1.add_projection(MappingProjection(sender=A, receiver=B), A, B)

        inputs_dict = {
            A: [[5.]],
        }


        comp2 = Composition(name="second_composition")

        A2 = TransferMechanism(name="A2",
                              function=Linear(slope=2.0))

        B2 = TransferMechanism(name="B2",
                              function=Linear(slope=3.0))

        comp2.add_node(A2)
        comp2.add_node(B2)

        comp2.add_projection(MappingProjection(sender=A2, receiver=B2), A2, B2)

        comp3 = Composition(name="outer_composition")
        comp3.add_node(comp1)
        comp3.add_node(comp2)
        comp3.add_projection(MappingProjection(), comp1, comp2)

        # comp1:
        # input = 5.0
        # mechA: 2.0*5.0 = 10.0
        # mechB: 3.0*10.0 = 30.0
        # output = 30.0

        # comp2:
        # input = 30.0
        # mechA2: 2.0*30.0 = 60.0
        # mechB2: 3.0*60.0 = 180.0
        # output = 180.0

        # comp3:
        # input = 5.0
        # output = 180.0
        comp3.run(inputs={comp1: [[5.]]}, execution_mode=comp_mode)
        np.testing.assert_allclose(comp3.results, [[[180.0]]])
        if comp_mode is pnl.ExecutionMode.Python:
            np.testing.assert_allclose(comp1.output_port.parameters.value.get(comp3), [30.0])
            np.testing.assert_allclose(comp2.output_port.parameters.value.get(comp3), [180.0])
            np.testing.assert_allclose(comp3.output_port.parameters.value.get(comp3), [180.0])

    @pytest.mark.nested
    @pytest.mark.composition
    def test_connect_compositions_with_complicated_states(self, comp_mode):

        inner_composition_1 = Composition(name="comp1")

        A = TransferMechanism(name="A1",
                              default_variable=[[0.0], [0.0]],
                              function=Linear(slope=2.0))

        B = TransferMechanism(name="B1",
                              default_variable=[[0.0], [0.0]],
                              function=Linear(slope=3.0))

        inner_composition_1.add_node(A)
        inner_composition_1.add_node(B)

        inner_composition_1.add_projection(MappingProjection(sender=A, receiver=B), A, B)
        inner_composition_1.add_projection(MappingProjection(sender=A.output_ports[1], receiver=B.input_ports[1]), A,
                                           B)

        inner_composition_2 = Composition(name="comp2")

        A2 = TransferMechanism(name="A2",
                              default_variable=[[0.0], [0.0]],
                              function=Linear(slope=2.0))

        B2 = TransferMechanism(name="B2",
                              default_variable=[[0.0], [0.0]],
                              function=Linear(slope=3.0))

        inner_composition_2.add_node(A2)
        inner_composition_2.add_node(B2)

        inner_composition_2.add_projection(MappingProjection(sender=A2, receiver=B2), A2, B2)
        inner_composition_2.add_projection(MappingProjection(sender=A2.output_ports[1], receiver=B2.input_ports[1]),
                                           A2, B2)

        outer_composition = Composition(name="outer_composition")

        outer_composition.add_node(inner_composition_1)
        outer_composition.add_node(inner_composition_2)

        outer_composition.add_projection(projection=MappingProjection(), sender=inner_composition_1,
                                         receiver=inner_composition_2)
        outer_composition.add_projection(
            projection=MappingProjection(sender=inner_composition_1.output_CIM.output_ports[1],
                                         receiver=inner_composition_2.input_CIM.input_ports[1]),
            sender=inner_composition_1, receiver=inner_composition_2)

        sched = Scheduler(composition=outer_composition)
        output = outer_composition.run(
            inputs={inner_composition_1: [[[5.0], [50.0]]]},
            scheduler=sched,
            execution_mode=comp_mode
        )

        np.testing.assert_allclose(output, [[180.], [1800.]])
        if comp_mode is pnl.ExecutionMode.Python:
            np.testing.assert_allclose(inner_composition_1.get_output_values(outer_composition), [[30.], [300.]])
            np.testing.assert_allclose(inner_composition_2.get_output_values(outer_composition), [[180.], [1800.]])
            np.testing.assert_allclose(outer_composition.get_output_values(outer_composition), [[180.], [1800.]])

    @pytest.mark.nested
    @pytest.mark.composition
    def test_compositions_as_origin_nodes(self, comp_mode):

        inner_composition_1 = Composition(name="inner_composition_1")

        A = TransferMechanism(name="A",
                              function=Linear(slope=0.5))

        B = TransferMechanism(name="B",
                              function=Linear(slope=2.0))

        C = TransferMechanism(name="C",
                              function=Linear(slope=3.0))

        inner_composition_1.add_node(A)
        inner_composition_1.add_node(B)
        inner_composition_1.add_node(C)

        inner_composition_1.add_projection(MappingProjection(), A, C)
        inner_composition_1.add_projection(MappingProjection(), B, C)

        inner_composition_2 = Composition(name="inner_composition_2")

        A2 = TransferMechanism(name="A2",
                               function=Linear(slope=0.25))

        B2 = TransferMechanism(name="B2",
                               function=Linear(slope=1.0))

        inner_composition_2.add_node(A2)
        inner_composition_2.add_node(B2)

        inner_composition_2.add_projection(MappingProjection(), A2, B2)

        mechanism_d = TransferMechanism(name="D",
                                        function=Linear(slope=3.0))

        outer_composition = Composition(name="outer_composition")

        outer_composition.add_node(inner_composition_1)
        outer_composition.add_node(inner_composition_2)
        outer_composition.add_node(mechanism_d)

        outer_composition.add_projection(projection=MappingProjection(),
                                         sender=inner_composition_1,
                                         receiver=mechanism_d)
        outer_composition.add_projection(projection=MappingProjection(),
                                         sender=inner_composition_2,
                                         receiver=mechanism_d)

        sched = Scheduler(composition=outer_composition)

        # FIX: order of InputPorts on inner composition 1 is not stable
        output = outer_composition.run(
            inputs={
                # inner_composition_1: [[2.0], [1.0]],
                inner_composition_1: {A: [2.0],
                                          B: [1.0]},
                inner_composition_2: [[12.0]]},
            scheduler=sched,
            execution_mode=comp_mode
        )
        np.testing.assert_allclose(output, [[36.]])

        if comp_mode is pnl.ExecutionMode.Python:
            np.testing.assert_allclose(A.get_output_values(outer_composition), [[1.0]])
            np.testing.assert_allclose(B.get_output_values(outer_composition), [[2.0]])
            np.testing.assert_allclose(C.get_output_values(outer_composition), [[9.0]])
            np.testing.assert_allclose(A2.get_output_values(outer_composition), [[3.0]])
            np.testing.assert_allclose(B2.get_output_values(outer_composition), [[3.0]])
            np.testing.assert_allclose(inner_composition_1.get_output_values(outer_composition), [[9.0]])
            np.testing.assert_allclose(inner_composition_2.get_output_values(outer_composition), [[3.0]])
            np.testing.assert_allclose(mechanism_d.get_output_values(outer_composition), [[36.0]])
            np.testing.assert_allclose(outer_composition.get_output_values(outer_composition), [[36.0]])

    @pytest.mark.nested
    @pytest.mark.composition
    def test_compositions_as_origin_nodes_multiple_trials(self, comp_mode):

        inner_composition_1 = Composition(name="inner_composition_1")

        A = TransferMechanism(name="A",
                              function=Linear(slope=0.5))

        B = TransferMechanism(name="B",
                              function=Linear(slope=2.0))

        C = TransferMechanism(name="C",
                              function=Linear(slope=3.0))

        inner_composition_1.add_node(A)
        inner_composition_1.add_node(B)
        inner_composition_1.add_node(C)

        inner_composition_1.add_projection(MappingProjection(), A, C)
        inner_composition_1.add_projection(MappingProjection(), B, C)

        inner_composition_2 = Composition(name="inner_composition_2")

        A2 = TransferMechanism(name="A2",
                               function=Linear(slope=0.25))

        B2 = TransferMechanism(name="B2",
                               function=Linear(slope=1.0))

        inner_composition_2.add_node(A2)
        inner_composition_2.add_node(B2)

        inner_composition_2.add_projection(MappingProjection(), A2, B2)

        mechanism_d = TransferMechanism(name="D",
                                        function=Linear(slope=3.0))

        outer_composition = Composition(name="outer_composition")

        outer_composition.add_node(inner_composition_1)


        outer_composition.add_node(inner_composition_2)


        outer_composition.add_node(mechanism_d)

        inner_composition_1._analyze_graph()

        outer_composition.add_projection(projection=MappingProjection(), sender=inner_composition_1,
                                         receiver=mechanism_d)



        inner_composition_2._analyze_graph()

        outer_composition.add_projection(projection=MappingProjection(), sender=inner_composition_2,
                                         receiver=mechanism_d)

        sched = Scheduler(composition=outer_composition)

        # FIX: order of InputPorts on inner composition 1 is not stable
        output = outer_composition.run(
            inputs={
                inner_composition_1: {A: [[2.0], [1.5], [2.5]],
                                      B: [[1.0], [1.5], [1.5]]},
                inner_composition_2: [[12.0], [11.5], [12.5]]},
            scheduler=sched,
            execution_mode=comp_mode
        )

        # trial 0:
        # inner composition 1 = (0.5*2.0 + 2.0*1.0) * 3.0 = 9.0
        # inner composition 2 = 0.25*12.0 = 3.0
        # outer composition = (3.0 + 9.0) * 3.0 = 36.0

        # trial 1:
        # inner composition 1 = (0.5*1.5 + 2.0*1.5) * 3.0 = 11.25
        # inner composition 2 = 0.25*11.5 = 2.875
        # outer composition = (2.875 + 11.25) * 3.0 = 42.375

        # trial 2:
        # inner composition 1 = (0.5*2.5 + 2.0*1.5) * 3.0 = 12.75
        # inner composition 2 = 0.25*12.5 = 3.125
        # outer composition = (3.125 + 12.75) * 3.0 = 47.625

        np.testing.assert_allclose(output, np.array([[47.625]]))

    def test_input_specification_multiple_nested_compositions(self):

        # level_0 composition --------------------------------- innermost composition
        level_0 = Composition(name="level_0")

        A0 = TransferMechanism(name="A0",
                               default_variable=[[0.], [0.]],
                               function=Linear(slope=1.))
        B0 = TransferMechanism(name="B0",
                               function=Linear(slope=2.))

        level_0.add_node(A0)
        level_0.add_node(B0)
        level_0.add_projection(MappingProjection(), A0, B0)
        level_0.add_projection(MappingProjection(sender=A0.output_ports[1], receiver=B0), A0, B0)

        # level_1 composition ---------------------------------
        level_1 = Composition(name="level_1")

        A1 = TransferMechanism(name="A1",
                              function=Linear(slope=1.))
        B1 = TransferMechanism(name="B1",
                              function=Linear(slope=2.))

        level_1.add_node(level_0)
        level_1.add_node(A1)
        level_1.add_node(B1)
        level_1.add_projection(MappingProjection(), level_0, B1)
        level_1.add_projection(MappingProjection(), A1, B1)

        # level_2 composition --------------------------------- outermost composition
        level_2 = Composition(name="level_2")

        A2 = TransferMechanism(name="A2",
                               input_shapes=2,
                               function=Linear(slope=1.))
        B2 = TransferMechanism(name="B2",
                               function=Linear(slope=2.))

        level_2.add_node(level_1)
        level_2.add_node(A2)
        level_2.add_node(B2)
        level_2.add_projection(MappingProjection(), level_1, B2)
        level_2.add_projection(MappingProjection(), A2, B2)

        sched = Scheduler(composition=level_2)

        # FIX: order of InputPorts in each inner composition (level_0 and level_1)
        level_2.run(inputs={A2: [[1.0, 2.0]],
                            level_1: {A1: [[1.0]],
                                      level_0: {A0: [[1.0], [2.0]]}}},
                    scheduler=sched)

        # level_0 output = 2.0 * (1.0 + 2.0) = 6.0
        np.testing.assert_allclose(level_0.get_output_values(level_2), [[6.0]])
        # level_1 output = 2.0 * (1.0 + 6.0) = 14.0
        np.testing.assert_allclose(level_1.get_output_values(level_2), [[14.0]])
        # level_2 output = 2.0 * (1.0 + 2.0 + 14.0) = 34.0
        np.testing.assert_allclose(level_2.get_output_values(level_2), [[34.0]])

    def test_error_on_custom_cim_ports(self):

        comp = Composition()
        mech = ProcessingMechanism()
        comp.add_node(mech)

        # NOTE: Adding ports to CIM from command line is currenlty disallowed
        # warning_msg = ('You are attempting to add custom ports to a CIM, which can result in unpredictable behavior '
        #                 'and is therefore recommended against. If suitable, you should instead add ports to the '
        #                r'mechanism\(s\) that project to or are projected to from the CIM.')
        error_msg = ('Adding ports to a CompositionInterfaceMechanism is not supported at this time; '
                     'these are handled automatically when a Composition is created.')
        with pytest.raises(CompositionError) as error_text:
            comp.input_CIM.add_ports(InputPort())
        assert error_msg in str(error_text.value)

        # with pytest.warns(None) as w:
        #     comp._analyze_graph()
        #     comp.run({mech: [[1]]})
        #
        # assert len(w) == 0

    def test_user_added_ports(self):

        comp = Composition()
        mech = ProcessingMechanism()
        comp.add_node(mech)
        # instantiate custom input and output ports
        inp = InputPort()
        out = OutputPort()

        # NOTE: Adding ports to CIM from command line is currenlty disallowed
        # # add custom input and output ports to CIM
        # comp.input_CIM.add_ports([inp, out])
        # # verify the ports have been added to the user_added_ports set
        # # and that no extra ports were added
        # assert inp in comp.input_CIM.user_added_ports['input_ports']
        # assert len(comp.input_CIM.user_added_ports['input_ports']) == 1
        # assert out in comp.input_CIM.user_added_ports['output_ports']
        # assert len(comp.input_CIM.user_added_ports['output_ports']) == 1
        # comp.input_CIM.remove_ports([inp, out])
        # # verify that call to remove ports succesfully removed the ports from user_added_ports
        # assert len(comp.input_CIM.user_added_ports['input_ports']) == 0
        # assert len(comp.input_CIM.user_added_ports['output_ports']) == 0

        error_msg = ('Adding ports to a CompositionInterfaceMechanism is not supported at this time; '
                     'these are handled automatically when a Composition is created.')
        with pytest.raises(CompositionError) as error_text:
            comp.input_CIM.add_ports([inp, out])
        assert error_msg in str(error_text.value)


    def test_parameter_CIM_port_order(self) -> object:
        # Note:  CIM_port order is also tested in TestNodes and test_simplified_necker_cube()

        # Inner Composition
        ia = TransferMechanism(name='ia')
        icomp = Composition(name='icomp', pathways=[ia])

        # Outer Composition
        ocomp = Composition(name='ocomp', pathways=[icomp])
        ocm = OptimizationControlMechanism(name='ic',
                                           agent_rep=ocomp,
                                           control_signals=[
                                               ControlSignal(projections=[(NOISE, ia)]),
                                               ControlSignal(projections=[(INTERCEPT, ia)]),
                                               ControlSignal(projections=[(SLOPE, ia)]),
                                           ],
                                           search_space=[[1], [1], [1]]
                                           )
        ocomp.add_controller(ocm)

        assert INTERCEPT in icomp.parameter_CIM.output_ports.names[0]
        assert NOISE in icomp.parameter_CIM.output_ports.names[1]
        assert SLOPE in icomp.parameter_CIM.output_ports.names[2]

    @pytest.fixture
    def parameter_CIM_routing_composition(self):
        # Inner Composition
        ia = TransferMechanism(name='ia')
        ib = TransferMechanism(name='ib')
        icomp = Composition(name='icomp', pathways=[ia, ib])
        # Outer Composition
        ocomp = Composition(name='ocomp')
        cm = ControlMechanism(
            name='control_mechanism',
            control_signals=ControlSignal(projections=[(SLOPE, ib)])
        )
        return ia, ib, cm, icomp, ocomp

    def test_parameter_CIM_routing_from_ControlMechanism_pathway_explicit(
        self, parameter_CIM_routing_composition
    ):
        ia, ib, cm, icomp, ocomp = parameter_CIM_routing_composition
        # warning_msg = f"A MappingProjection has been created from a ControlSignal of 'control_mechanism' " \
        #               f"-- specified in 'pathway' arg for add_linear_processing_pathway method of 'ocomp' -- " \
        #               f"to another Mechanism in that pathway.  " \
        #               f"If this is not the intended behavior, add 'control_mechanism' separately to 'ocomp'."
        # with pytest.warns(UserWarning, match=warning_msg):
        ocomp.add_linear_processing_pathway([cm, icomp])
            # ocomp.add_linear_processing_pathway([icomp])
            # ocomp.add_node(cm)
            # ocomp.show_graph(show_cim=True, show_node_structure=True)
        ocomp._analyze_graph()
        input_nodes = ocomp.get_nodes_by_role(NodeRole.INPUT)
        assert cm in input_nodes
        assert icomp in input_nodes

        ocomp.run(
            inputs={
                cm: [[2], [2], [2]],
                icomp: [[2], [2], [2]],
            }
        )
        # Input of 2's multiplied by the controlled slope of ib (2)
        #   (since control_mechanism executes before ib) results in 4's
        np.testing.assert_array_almost_equal(ocomp.results, [[[4]], [[4]], [[4]]])
        assert len(ib.mod_afferents) == 1
        assert len(cm.control_signals[0].efferents) == 1
        assert cm.control_signals[0].efferents[0].receiver.owner == icomp.parameter_CIM
        assert ib.mod_afferents[0].sender == icomp.parameter_CIM.output_port
        assert icomp.parameter_CIM_ports[ib.parameter_ports['slope']][0].path_afferents[0].sender == cm.output_port
        assert cm in ocomp.graph_processing.dependency_dict[icomp]

    def test_parameter_CIM_routing_from_ControlMechanism_pathway_implicit(
        self, parameter_CIM_routing_composition
    ):
        ia, ib, cm, icomp, ocomp = parameter_CIM_routing_composition
        ocomp.add_node(cm)
        ocomp.add_node(icomp)

        ocomp._analyze_graph()
        input_nodes = ocomp.get_nodes_by_role(NodeRole.INPUT)
        assert cm in input_nodes
        assert icomp in input_nodes

        ocomp.run(
            inputs={
                cm: [[2], [2], [2]],
                icomp: [[2], [2], [2]],
            }
        )
        # first trial result for icomp is 2 because cm and icomp run
        # simultaneously, so the value of cm control signals is still
        # cm's default (1) and not its input (2)
        np.testing.assert_allclose(ocomp.results, [[[2]], [[4]], [[4]]])
        assert len(ib.mod_afferents) == 1
        assert ib.mod_afferents[0].sender == icomp.parameter_CIM.output_port
        assert icomp.parameter_CIM_ports[ib.parameter_ports['slope']][0].path_afferents[0].sender == cm.output_port
        # Verify that only Projections from cm are to parameter_CIM (and not to icomp.input_CIM)
        assert all([proj.receiver.owner == icomp.parameter_CIM for proj in cm.efferents])
        assert cm not in ocomp.graph_processing.dependency_dict[icomp]

    def test_nested_control_projection_count_controller(self):
        # Inner Composition
        ia = TransferMechanism(name='ia')
        icomp = Composition(name='icomp', pathways=[ia])
        # Outer Composition
        ocomp = Composition(name='ocomp', pathways=[icomp])
        ocm = OptimizationControlMechanism(name='ocm',
                                           agent_rep=ocomp,
                                           control_signals=[
                                               ControlSignal(projections=[(NOISE, ia)]),
                                               ControlSignal(projections=[(INTERCEPT, ia)]),
                                               ControlSignal(projections=[(SLOPE, ia)]),
                                           ],
                                           search_space=[
                                               [1], [1], [1]
                                           ]
                                           )
        ocomp.add_controller(ocm)
        assert len(ocm.efferents) == 3
        assert all([proj.receiver.owner == icomp.parameter_CIM for proj in ocm.efferents])
        assert len(ia.mod_afferents) == 3
        assert all([proj.sender.owner == icomp.parameter_CIM for proj in ia.mod_afferents])

    def test_nested_control_projection_count_control_mech(self):
        # Inner Composition
        ia = TransferMechanism(name='ia')
        icomp = Composition(name='icomp', pathways=[ia])
        # Outer Composition
        oa = TransferMechanism(name='oa')
        cm = ControlMechanism(name='cm',
            control=[
            ControlSignal(projections=[(NOISE, ia)]),
            ControlSignal(projections=[(INTERCEPT, ia)]),
            ControlSignal(projections=[(SLOPE, ia)])
            ]
        )
        ocomp = Composition(name='ocomp', pathways=[[oa, icomp], [cm]])
        assert len(cm.efferents) == 3
        assert all([proj.receiver.owner == icomp.parameter_CIM for proj in cm.efferents])
        assert len(ia.mod_afferents) == 3
        assert all([proj.sender.owner == icomp.parameter_CIM for proj in ia.mod_afferents])


class TestInputCIMOutputPortToOriginOneToMany:

    def test_one_to_two(self):
        A = ProcessingMechanism(name='A')
        B = ProcessingMechanism(name='B')
        C = ProcessingMechanism(name='C',
                                input_ports=[A.input_port])

        comp = Composition(name='comp')

        comp.add_linear_processing_pathway([A, B])
        comp.add_node(C)

        comp.run(inputs={A: [[1.23]]})

        np.testing.assert_allclose(A.parameters.value.get(comp), [[1.23]])
        np.testing.assert_allclose(B.parameters.value.get(comp), [[1.23]])
        np.testing.assert_allclose(C.parameters.value.get(comp), [[1.23]])

    def test_origin_input_source_true_no_input(self):
        A = ProcessingMechanism(name='A')
        B = ProcessingMechanism(name='B')
        C = ProcessingMechanism(name='C',
                                default_variable=[[4.56]])

        comp = Composition(name='comp')

        comp.add_linear_processing_pathway([A, B])
        comp.add_node(C)

        comp.run(inputs={A: [[1.23]]})

        np.testing.assert_allclose(A.parameters.value.get(comp), [[1.23]])
        np.testing.assert_allclose(B.parameters.value.get(comp), [[1.23]])
        np.testing.assert_allclose(C.parameters.value.get(comp), [[4.56]])

    def test_mix_and_match_input_sources(self):
        A = ProcessingMechanism(name='A')
        B = ProcessingMechanism(name='B',
                                default_variable=[[0.], [0.]])
        C = ProcessingMechanism(name='C',
                                input_ports=[B.input_ports[1], A.input_port, B.input_ports[0]])

        input_dict = {A: [[2.0]],
                      B: [[3.0], [1.0]]}

        comp = Composition(name="comp")

        comp.add_node(A)
        comp.add_node(B)
        comp.add_node(C)

        comp.run(inputs=input_dict)

        np.testing.assert_allclose(A.parameters.value.get(comp), [[2.]])
        np.testing.assert_allclose(B.parameters.value.get(comp), [[3.], [1.]])
        np.testing.assert_allclose(C.parameters.value.get(comp), [[1.], [2.], [3.]])

    def test_non_origin_partial_input_spec(self):
        A = ProcessingMechanism(name='A',
                                function=Linear(slope=2.0))
        B = ProcessingMechanism(name='B',
                                input_ports=[[0.], A.input_port])

        comp = Composition(name='comp')

        comp.add_linear_processing_pathway([A, B])

        comp.run(inputs={A: [[1.23]]})
        np.testing.assert_allclose(B.get_input_values(comp), [[2.46], [1.23]])

class TestInputSpec:

    def test_valid_mismatched_input_lens(self):
        A = ProcessingMechanism(name="A")
        B = ProcessingMechanism(name="B")
        C = ProcessingMechanism(name="C")

        comp = Composition(name="COMP")

        comp.add_linear_processing_pathway([A, C])
        comp.add_linear_processing_pathway([B, C])

        inputs_to_A = [[1.0]]                           # same (1.0) on every trial
        inputs_to_B = [[1.0], [2.0], [3.0], [4.0]]      # increment on every trial

        results_A = []
        results_B = []
        results_C = []

        def call_after_trial():
            results_A.append(A.parameters.value.get(comp))
            results_B.append(B.parameters.value.get(comp))
            results_C.append(C.parameters.value.get(comp))

        comp.run(inputs={A: inputs_to_A,
                         B: inputs_to_B},
                 call_after_trial=call_after_trial)

        np.testing.assert_allclose(results_A, [[[1.0]], [[1.0]], [[1.0]], [[1.0]]])
        np.testing.assert_allclose(results_B, [[[1.0]], [[2.0]], [[3.0]], [[4.0]]])
        np.testing.assert_allclose(results_C, [[[2.0]], [[3.0]], [[4.0]], [[5.0]]])

    def test_valid_only_one_node_provides_input_spec(self):
        A = ProcessingMechanism(name="A",
                                default_variable=[[1.5]])   # default variable will be used as input to this INPUT node
        B = ProcessingMechanism(name="B")
        C = ProcessingMechanism(name="C")

        comp = Composition(name="COMP")

        comp.add_linear_processing_pathway([A, C])
        comp.add_linear_processing_pathway([B, C])

        inputs_to_B = [[1.0], [2.0], [3.0], [4.0]]      # increment on every trial

        results_A = []
        results_B = []
        results_C = []

        def call_after_trial():
            results_A.append(A.parameters.value.get(comp))
            results_B.append(B.parameters.value.get(comp))
            results_C.append(C.parameters.value.get(comp))

        comp.run(inputs={B: inputs_to_B},
                 call_after_trial=call_after_trial)

        np.testing.assert_allclose(results_A, [[[1.5]], [[1.5]], [[1.5]], [[1.5]]])
        np.testing.assert_allclose(results_B, [[[1.0]], [[2.0]], [[3.0]], [[4.0]]])
        np.testing.assert_allclose(results_C, [[[2.5]], [[3.5]], [[4.5]], [[5.5]]])

    def test_invalid_mismatched_input_lens(self):
        A = ProcessingMechanism(name="A")
        B = ProcessingMechanism(name="B")
        C = ProcessingMechanism(name="C")

        comp = Composition(name="COMP")

        comp.add_linear_processing_pathway([A, C])
        comp.add_linear_processing_pathway([B, C])

        inputs_to_A = [[1.0], [2.0]]                    # 2 input specs
        inputs_to_B = [[1.0], [2.0], [3.0], [4.0]]      # 4 input specs

        with pytest.raises(CompositionError) as error_text:
            comp.run(inputs={A: inputs_to_A,
                             B: inputs_to_B})
        assert "input dictionary for COMP contains input specifications of different lengths" in str(error_text.value)

    def test_valid_input_float(self):
        A = ProcessingMechanism(name="A")
        comp = Composition(name="comp")
        comp.add_node(A)

        comp.run(inputs={A: 5.0})
        np.testing.assert_allclose(comp.results, [[[5.0]]])

        comp.run(inputs={A: [5.0, 10.0, 15.0]})
        np.testing.assert_allclose(comp.results, [[[5.0]], [[5.0]], [[10.0]], [[15.0]]])


class TestSimplifedNestedCompositionSyntax:
    def test_connect_outer_composition_to_only_input_node_in_inner_comp_option1(self):

        inner1 = Composition(name="inner")

        A1 = TransferMechanism(name="A1",
                               function=Linear(slope=2.0))

        B1 = TransferMechanism(name="B1",
                               function=Linear(slope=3.0))

        inner1.add_linear_processing_pathway([A1, B1])

        inner2 = Composition(name="inner2")

        A2 = TransferMechanism(name="A2",
                               function=Linear(slope=2.0))

        B2 = TransferMechanism(name="B2",
                               function=Linear(slope=3.0))

        inner2.add_linear_processing_pathway([A2, B2])

        outer = Composition(name="outer")
        outer.add_nodes([inner1, inner2])
        outer.add_projection(sender=B1, receiver=inner2)

        # comp1:  input = 5.0   |  mechA: 2.0*5.0 = 10.0    |  mechB: 3.0*10.0 = 30.0    |  output = 30.0
        # comp2:  input = 30.0  |  mechA2: 2.0*30.0 = 60.0  |  mechB2: 3.0*60.0 = 180.0  |  output = 180.0
        # comp3:  input = 5.0   |  output = 180.0

        res = outer.run(inputs={inner1: [[5.]]})
        np.testing.assert_allclose(res, [[180.0]])

        np.testing.assert_allclose(inner1.output_port.parameters.value.get(outer), [30.0])
        np.testing.assert_allclose(inner2.output_port.parameters.value.get(outer), [180.0])
        np.testing.assert_allclose(outer.output_port.parameters.value.get(outer), [180.0])

    def test_connect_outer_composition_to_only_input_node_in_inner_comp_option2(self):
        inner1 = Composition(name="inner")

        A1 = TransferMechanism(name="A1",
                               function=Linear(slope=2.0))

        B1 = TransferMechanism(name="B1",
                               function=Linear(slope=3.0))

        inner1.add_linear_processing_pathway([A1, B1])

        inner2 = Composition(name="inner2")

        A2 = TransferMechanism(name="A2",
                               function=Linear(slope=2.0))

        B2 = TransferMechanism(name="B2",
                               function=Linear(slope=3.0))

        inner2.add_linear_processing_pathway([A2, B2])

        outer = Composition(name="outer")
        outer.add_nodes([inner1, inner2])
        outer.add_projection(sender=inner1, receiver=A2)

        # CRASHING WITH: FIX 6/1/20
        # subprocess.CalledProcessError: Command '['dot', '-Tpdf', '-O', 'outer']' returned non-zero exit status 1.
        # outer.show_graph(show_node_structure=True,
        #                  show_nested=True)

        # comp1:  input = 5.0   |  mechA: 2.0*5.0 = 10.0    |  mechB: 3.0*10.0 = 30.0    |  output = 30.0
        # comp2:  input = 30.0  |  mechA2: 2.0*30.0 = 60.0  |  mechB2: 3.0*60.0 = 180.0  |  output = 180.0
        # comp3:  input = 5.0   |  output = 180.0

        res = outer.run(inputs={inner1: [[5.]]})
        np.testing.assert_allclose(res, [[180.0]])

        np.testing.assert_allclose(inner1.output_port.parameters.value.get(outer), [30.0])
        np.testing.assert_allclose(inner2.output_port.parameters.value.get(outer), [180.0])
        np.testing.assert_allclose(outer.output_port.parameters.value.get(outer), [180.0])

    def test_connect_outer_composition_to_only_input_node_in_inner_comp_option3(self):

        inner1 = Composition(name="inner")

        A1 = TransferMechanism(name="A1",
                               function=Linear(slope=2.0))

        B1 = TransferMechanism(name="B1",
                               function=Linear(slope=3.0))

        inner1.add_linear_processing_pathway([A1, B1])

        inner2 = Composition(name="inner2")

        A2 = TransferMechanism(name="A2",
                               function=Linear(slope=2.0))

        B2 = TransferMechanism(name="B2",
                               function=Linear(slope=3.0))

        inner2.add_linear_processing_pathway([A2, B2])

        outer = Composition(name="outer")
        outer.add_nodes([inner1, inner2])
        outer.add_projection(sender=B1, receiver=A2)

        # comp1:  input = 5.0   |  mechA: 2.0*5.0 = 10.0    |  mechB: 3.0*10.0 = 30.0    |  output = 30.0
        # comp2:  input = 30.0  |  mechA2: 2.0*30.0 = 60.0  |  mechB2: 3.0*60.0 = 180.0  |  output = 180.0
        # comp3:  input = 5.0   |  output = 180.0

        res = outer.run(inputs={inner1: [[5.]]})
        np.testing.assert_allclose(res, [[180.0]])

        np.testing.assert_allclose(inner1.output_port.parameters.value.get(outer), [30.0])
        np.testing.assert_allclose(inner2.output_port.parameters.value.get(outer), [180.0])
        np.testing.assert_allclose(outer.output_port.parameters.value.get(outer), [180.0])

    def test_connect_outer_composition_to_all_input_nodes_in_inner_comp(self):

        inner1 = Composition(name="inner")
        A1 = TransferMechanism(name="A1",
                               function=Linear(slope=2.0))
        B1 = TransferMechanism(name="B1",
                               function=Linear(slope=3.0))

        inner1.add_linear_processing_pathway([A1, B1])

        inner2 = Composition(name="inner2")
        A2 = TransferMechanism(name="A2")
        B2 = TransferMechanism(name="B2")
        C2 = TransferMechanism(name="C2")

        inner2.add_nodes([A2, B2, C2])

        outer1 = Composition(name="outer1")
        outer1.add_nodes([inner1, inner2])

        # Spec 1: add projection *node in* inner1 --> inner 2 (implies first InputPort -- corresponding to A2)
        outer1.add_projection(sender=B1, receiver=inner2)
        # Spec 2:  add projection *node in* inner1 --> *node in* inner2
        outer1.add_projection(sender=B1, receiver=B2)
        # Spec 3: add projection inner1 --> *node in* inner2
        outer1.add_projection(sender=inner1, receiver=C2)
        eid = "eid"
        outer1.run(inputs={inner1: [[1.]]},
                   context=eid)

        np.testing.assert_allclose(A1.parameters.value.get(eid), [[2.0]])
        np.testing.assert_allclose(B1.parameters.value.get(eid), [[6.0]])

        for node in [A2, B2, C2]:
            np.testing.assert_allclose(node.parameters.value.get(eid), [[6.0]])

@pytest.mark.parametrize(
    # expected_input_shape: one input per source input_port
    # expected_output_shape: one output per terminal output_port
    # TODO: change mechanisms to input_ports and output_ports args?
    'mechanisms, expected_input_shape, expected_output_shape',
    [
        (
            [TransferMechanism()], [[0]], [[0]]
        ),
        (
            [
                TransferMechanism(),
                TransferMechanism(output_ports=['RESULT', 'MEAN']),
            ],
            [[0], [0]],
            [[0], [0], [0]],
        ),
        (
            [
                TransferMechanism(),
                TransferMechanism(output_ports=['RESULT', 'MEAN']),
                TransferMechanism(output_ports=['RESULT', 'MEAN', 'MEDIAN']),
            ],
            [[0], [0], [0]],
            [[0], [0], [0], [0], [0], [0]],
        ),
        (
            [
                TransferMechanism(input_ports=['Port1', 'Port2']),
                TransferMechanism(output_ports=['RESULT', 'MEAN']),
            ],
            [[0], [0], [0]],
            # an output_port generated for each custom input_port
            # overwrites output_ports arg - intentional?
            [[0], [0], [0], [0]],
        ),
        (
            [
                TransferMechanism(),
                TransferMechanism(
                    input_ports=['Port1', 'Port2'],
                ),
                TransferMechanism(
                    input_ports=['Port1', 'Port2', 'Port3'],
                ),
            ],
            [[0], [0], [0], [0], [0], [0]],
            [[0], [0], [0], [0], [0], [0]],
        ),
    ]
)
def test_CIM_shapes(mechanisms, expected_input_shape, expected_output_shape):
    comp = Composition()

    for i in range(len(mechanisms)):
        comp.add_node(mechanisms[i])

    comp._analyze_graph()

    for cim, expected_shape in [
        (comp.input_CIM, expected_input_shape),
        (comp.output_CIM, expected_output_shape),
    ]:
        np.testing.assert_array_equal(
            cim.defaults.variable.shape,
            cim.defaults.value.shape,
            err_msg=f'{cim}:',
            verbose=True
        )
        np.testing.assert_array_equal(
            cim.defaults.variable.shape,
            cim.parameters.variable.get().shape,
            err_msg=f'{cim}:',
            verbose=True
        )
        np.testing.assert_array_equal(
            cim.defaults.value.shape,
            cim.parameters.value.get().shape,
            err_msg=f'{cim}:',
            verbose=True
        )
        np.testing.assert_array_equal(
            cim.defaults.variable,
            convert_all_elements_to_np_array(expected_shape),
            err_msg=f'{cim}:',
            verbose=True
        )



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* Time ***************************************************************
import graph_scheduler

__all__ = [
    'TimeScale', 'Time', 'TimeHistoryTree'
]

TimeScale = graph_scheduler.TimeScale
Time = graph_scheduler.Time
TimeHistoryTree = graph_scheduler.TimeHistoryTree

_doc_subs = {
    'TimeScale': [
        (
            r'(a single\n *call to `run <Scheduler\.run>`)',
            '\\1 (a single input to a\n        `Composition <Composition>`.)'
        ),
        (
            r', managed by the\n *environment using the Scheduler.',
            r'. In PsyNeuLink, this is the scope of a call to the `run <Composition.run>` method of a `Composition <Composition>`, consisting of one or more `TRIALs <TimeScale.TRIAL>`.'
        ),
    ]
}



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

# *********************************************  ComparatorMechanism ***************************************************

"""

Contents
--------

  * `ComparatorMechanism_Overview`
  * `ComparatorMechanism_Creation`
  * `ComparatorMechanism_Structure`
  * `ComparatorMechanism_Execution`
  * `ComparatorMechanism_Example`
  * `ComparatorMechanism_Class_Reference`


.. _ComparatorMechanism_Overview:

Overview
--------

A ComparatorMechanism is a subclass of `ObjectiveMechanism` that receives two inputs (a sample and a target), compares
them using its `function <ComparatorMechanism.function>`, and places the calculated discrepancy between the two in its
*OUTCOME* `OutputPort <ComparatorMechanism.output_port>`.

.. _ComparatorMechanism_Creation:

Creating a ComparatorMechanism
------------------------------

ComparatorMechanisms are generally created automatically when other PsyNeuLink components are created (such as
`LearningMechanisms <LearningMechanism_Creation>`).  A ComparatorMechanism can also be created directly by calling
its constructor.  Its **sample** and **target** arguments are used to specify the OutputPorts that provide the
sample and target inputs, respectively (see `ObjectiveMechanism_Monitored_ports` for details concerning their
specification, which are special versions of an ObjectiveMechanism's **monitor** argument).  When the
ComparatorMechanism is created, two InputPorts are created, one each for its sample and target inputs (and named,
by default, *SAMPLE* and *TARGET*). Each is assigned a MappingProjection from the corresponding OutputPort specified
in the **sample** and **target** arguments.

It is important to recognize that the value of the *SAMPLE* and *TARGET* InputPorts must have the same length and type,
so that they can be compared using the ComparatorMechanism's `function <ComparatorMechanism.function>`.  By default,
they use the format of the OutputPorts specified in the **sample** and **target** arguments, respectively,
and the `MappingProjection` to each uses an `IDENTITY_MATRIX`.  Therefore, for the default configuration, the
OutputPorts specified in the **sample** and **target** arguments must have values of the same length and type.
If these differ, the **input_ports** argument can be used to explicitly specify the format of the ComparatorMechanism's
*SAMPLE* and *TARGET* InputPorts, to insure they are compatible with one another (as well as to customize their
names, if desired).  If the **input_ports** argument is used, *both* the sample and target InputPorts must be
specified.  Any of the formats for `specifying InputPorts <InputPort_Specification>` can be used in the argument.
If values are assigned for the InputPorts, they must be of equal length and type.  Their types must
also be compatible with the value of the OutputPorts specified in the **sample** and **target** arguments.  However,
the length specified for an InputPort can differ from its corresponding OutputPort;  in that case, by default, the
MappingProjection created uses a `FULL_CONNECTIVITY` matrix.  Thus, OutputPorts of differing lengths can be mapped
to the sample and target InputPorts of a ComparatorMechanism (see the `example <ComparatorMechanism_Example>` below),
so long as the latter are of the same length.  If a projection other than a `FULL_CONNECTIVITY` matrix is needed, this
can be specified using the *PROJECTION* entry of a `Port specification dictionary <Port_Specification>` for the
InputPort in the **input_ports** argument.

.. _ComparatorMechanism_Structure:

Structure
---------

A ComparatorMechanism has two `input_ports <ComparatorMechanism.input_ports>`, each of which receives a
`MappingProjection` from a corresponding OutputPort specified in the **sample** and **target** arguments of its
constructor.  The InputPorts are listed in the Mechanism's `input_ports <ComparatorMechanism.input_ports>` attribute
and named, respectively, *SAMPLE* and *TARGET*.  The OutputPorts from which they receive their projections (specified
in the **sample** and **target** arguments) are listed in the Mechanism's `sample <ComparatorMechanism.sample>` and
`target <ComparatorMechanism.target>` attributes as well as in its `monitor <ComparatorMechanism.monitor>` attribute.
The ComparatorMechanism's `function <ComparatorMechanism.function>` compares the value of the sample and target
InputPorts.  By default, it uses a `LinearCombination` function, assigning the sample InputPort a `weight
<LinearCombination.weight>` of *-1* and the target a `weight <LinearCombination.weight>` of *1*, so that the sample
is subtracted from the target.  However, the `function <ComparatorMechanism.function>` can be customized, so long as
it is replaced with one that takes two arrays with the same format as its inputs and generates a similar array as its
result. The result is assigned as the value of the Comparator Mechanism's *OUTCOME* (`primary <OutputPort_Primary>`)
OutputPort.

.. _ComparatorMechanism_Execution:

Execution
---------

When a ComparatorMechanism is executed, it updates its input_ports with the values of the OutputPorts specified
in its **sample** and **target** arguments, and then uses its `function <ComparatorMechanism.function>` to
compare these.  By default, the result is assigned to the `value <Mechanism_Base.value>` of its *OUTCOME*
`output_port <ComparatorMechanism.output_port>`, and as the first item of the Mechanism's
`output_values <ComparatorMechanism.output_values>` attribute.

.. _ComparatorMechanism_Example:

Example
-------

.. _ComparatorMechanism_Default_Input_Value_Example:

*Formatting InputPort values*

The **default_variable** argument can be used to specify a particular format for the SAMPLE and/or TARGET InputPorts
of a ComparatorMechanism.  This can be useful when one or both of these differ from the format of the
OutputPort(s) specified in the **sample** and **target** arguments. For example, for `Reinforcement Learning
<Reinforcement>`, a ComparatorMechanism is used to monitor an action selection Mechanism (the sample), and compare
this with a reinforcement signal (the target).  In the example below, the action selection Mechanism is a
`TransferMechanism` that uses the `SoftMax` function (and the `PROB <Softmax.PROB>` as its output format) to select
an action.  This generates a vector with a single non-zero value (the selected action). Because the output is a vector,
specifying it as the ComparatorMechanism's **sample** argument will generate a corresponding InputPort with a vector
as its value.  This will not match the reward signal specified in the ComparatorMechanism's **target** argument, the
value of which is a single scalar.  This can be dealt with by explicitly specifying the format for the SAMPLE and
TARGET InputPorts in the **default_variable** argument of the ComparatorMechanism's constructor, as follows::

    >>> import psyneulink as pnl
    >>> my_action_selection_mech = pnl.TransferMechanism(input_shapes=5,
    ...                                                  function=pnl.SoftMax(output=pnl.PROB))

    >>> my_reward_mech = pnl.TransferMechanism()

    >>> my_comparator_mech = pnl.ComparatorMechanism(default_variable = [[0],[0]],
    ...                                              sample=my_action_selection_mech,
    ...                                              target=my_reward_mech)

Note that ``my_action_selection_mechanism`` is specified to take an array of length 5 as its input, and therefore
generate one of the same length as its `primary output <OutputPort_Primary>`.  Since it is assigned as the **sample**
of the ComparatorMechanism, by default this will create a *SAMPLE* InputPort of length 5, that will not match the
length of the *TARGET* InputPort (the default for which is length 1).  This is taken care of, by specifying the
**default_variable** argument as an array with two single-value arrays (corresponding to the *SAMPLE* and *TARGET*
InputPorts). (In this example, the **sample** and **target** arguments are specified as Mechanisms since,
by default, each has only a single (`primary <OutputPort_Primary>`) OutputPort, that will be used;  if either had
more than one OutputPort, and one of those was desired, it would have had to be specified explicitly in the
**sample** or **target** argument).

.. _ComparatorMechanism_Class_Reference:

Class Reference
---------------

"""

from collections.abc import Iterable

import numpy as np
from beartype import beartype

from psyneulink._typing import Optional, Union

from psyneulink.core.components.functions.nonstateful.transformfunctions import LinearCombination
from psyneulink.core.components.mechanisms.mechanism import Mechanism_Base, MechanismError
from psyneulink.core.components.mechanisms.processing.objectivemechanism import ObjectiveMechanism
from psyneulink.core.components.shellclasses import Mechanism
from psyneulink.core.components.ports.inputport import InputPort
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.components.ports.port import _parse_port_spec
from psyneulink.core.globals.keywords import \
    COMPARATOR_MECHANISM, FUNCTION, INPUT_PORTS, NAME, OUTCOME, SAMPLE, TARGET, \
    VARIABLE, PREFERENCE_SET_NAME, Loss, SUM
from psyneulink.core.globals.parameters import Parameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet, REPORT_OUTPUT_PREF
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel
from psyneulink.core.globals.utilities import \
    NumericCollections, is_value_spec, iscompatible, kwCompatibilityLength, kwCompatibilityNumeric, recursive_update
from psyneulink.core.globals.utilities import safe_len

__all__ = ['ComparatorMechanism', 'ComparatorMechanismError']

class ComparatorMechanismError(MechanismError):
    pass


class ComparatorMechanism(ObjectiveMechanism):
    """
    ComparatorMechanism(                                \
        sample,                                         \
        target,                                         \
        input_ports=[SAMPLE,TARGET]                     \
        function=LinearCombination(weights=[[-1],[1]],  \
        output_ports=OUTCOME)

    Subclass of `ObjectiveMechanism` that compares the values of two `OutputPorts <OutputPort>`.
    See `ObjectiveMechanism <ObjectiveMechanism_Class_Reference>` for additional arguments and attributes.


    Arguments
    ---------

    sample : OutputPort, Mechanism, value, or string
        specifies the value to compare with the `target` by the `function <ComparatorMechanism.function>`.

    target :  OutputPort, Mechanism, value, or string
        specifies the value with which the `sample` is compared by the `function <ComparatorMechanism.function>`.

    input_ports :  List[InputPort, value, str or dict] or Dict[] : default [SAMPLE, TARGET]
        specifies the names and/or formats to use for the values of the sample and target InputPorts;
        by default they are named *SAMPLE* and *TARGET*, and their formats are match the value of the OutputPorts
        specified in the **sample** and **target** arguments, respectively (see `ComparatorMechanism_Structure`
        for additional details).

    function :  Function, function or method : default Distance(metric=DIFFERENCE)
        specifies the `function <Comparator.function>` used to compare the `sample` with the `target`.


    Attributes
    ----------

    COMMENT:
    default_variable : Optional[List[array] or 2d np.array]
    COMMENT

    sample : OutputPort
        determines the value to compare with the `target` by the `function <ComparatorMechanism.function>`.

    target : OutputPort
        determines the value with which `sample` is compared by the `function <ComparatorMechanism.function>`.

    input_ports : ContentAddressableList[InputPort, InputPort]
        contains the two InputPorts named, by default, *SAMPLE* and *TARGET*, each of which receives a
        `MappingProjection` from the OutputPorts referenced by the `sample` and `target` attributes
        (see `ComparatorMechanism_Structure` for additional details).

    function : TransformFunction, function or method
        used to compare the `sample` with the `target`.  It can be any `TransformFunction <Transformfunctions>`,
        or a python function that takes a 2d array with two items and returns a 1d array of the same length as the
        two input items.

    output_port : OutputPort
        contains the `primary <OutputPort_Primary>` OutputPort of the ComparatorMechanism; the default is
        its *OUTCOME* OutputPort, the value of which is equal to the `value <ComparatorMechanism.value>`
        attribute of the ComparatorMechanism.

    output_ports : ContentAddressableList[OutputPort]
        contains, by default, only the *OUTCOME* (primary) OutputPort of the ComparatorMechanism.

    output_values : 2d np.array
        contains one item that is the value of the *OUTCOME* OutputPort.

    standard_output_ports : list[str]
        list of `Standard OutputPorts <OutputPort_Standard>` that includes the following in addition to the
        `standard_output_ports <ObjectiveMechanism.standard_output_ports>` of an `ObjectiveMechanism`:

        .. _COMPARATOR_MECHANISM_SSE

        *SUM*
            the sum of the terms in in the array returned by the Mechanism's function.

        *SSE*
            the sum of squares of the terms in the array returned by the Mechanism's function.

        .. _COMPARATOR_MECHANISM_MSE

        *MSE*
            the mean of the squares of the terms returned by the Mechanism's function.

        .. _COMPARATOR_MECHANISM_MSE


    """
    componentType = COMPARATOR_MECHANISM

    classPreferenceLevel = PreferenceLevel.SUBTYPE
    # These will override those specified in TYPE_DEFAULT_PREFERENCES
    classPreferences = {
        PREFERENCE_SET_NAME: 'ComparatorCustomClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE)}

    class Parameters(ObjectiveMechanism.Parameters):
        """
            Attributes
            ----------

                variable
                    see `variable <ComparatorMechanism.variable>`

                    :default value: numpy.array([[0], [0]])
                    :type: ``numpy.ndarray``
                    :read only: True

                function
                    see `function <ComparatorMechanism.function>`

                    :default value: `LinearCombination`(weights=numpy.array([[-1], [ 1]]))
                    :type: `Function`

                output_ports
                    see `output_ports <ComparatorMechanism.output_ports>`

                    :default value: [`OUTCOME`]
                    :type: ``list``
                    :read only: True

                sample
                    see `sample <ComparatorMechanism.sample>`

                    :default value: None
                    :type:

                target
                    see `target <ComparatorMechanism.target>`

                    :default value: None
                    :type:
        """
        # By default, ComparatorMechanism compares two 1D np.array input_ports
        variable = Parameter(np.array([[0], [0]]), read_only=True, pnl_internal=True, constructor_argument='default_variable')
        function = Parameter(LinearCombination(weights=[[-1], [1]]), stateful=False, loggable=False)
        sample = None
        target = None

        output_ports = Parameter(
            [OUTCOME],
            stateful=False,
            loggable=False,
            read_only=True,
            structural=True,
        )

    # ComparatorMechanism parameter and control signal assignments):

    standard_output_ports = ObjectiveMechanism.standard_output_ports.copy()
    standard_output_ports.extend([{NAME: SUM.upper(),
                                   FUNCTION: lambda x: np.sum(x)},
                                  {NAME: Loss.SSE.name,
                                   FUNCTION: lambda x: np.sum(x * x)},
                                  {NAME: Loss.MSE.name,
                                   FUNCTION: lambda x: np.sum(x * x) / safe_len(x)}]
                                 )
    standard_output_port_names = ObjectiveMechanism.standard_output_port_names.copy()
    standard_output_port_names.extend([SUM.upper(), Loss.SSE.name, Loss.MSE.name])

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable=None,
                 sample: Optional[Union[OutputPort, Mechanism_Base, dict, NumericCollections, str]] = None,
                 target: Optional[Union[OutputPort, Mechanism_Base, dict, NumericCollections, str]] = None,
                 function=None,
                 output_ports:Optional[Union[str, Iterable]] = None,
                 params=None,
                 name=None,
                 prefs:   Optional[ValidPrefSet] = None,
                 **kwargs
                 ):

        input_ports = kwargs.pop(INPUT_PORTS, {})
        input_ports_spec = input_ports
        if input_ports:
            input_ports = {INPUT_PORTS: input_ports}

        input_ports = self._merge_legacy_constructor_args(sample, target, default_variable, input_ports)

        # Default output_ports is specified in constructor as a tuple rather than a list
        # to avoid "gotcha" associated with mutable default arguments
        # (see: bit.ly/2uID3s3 and http://docs.python-guide.org/en/latest/writing/gotchas/)
        if isinstance(output_ports, (str, tuple)):
            output_ports = list(output_ports)

        # IMPLEMENTATION NOTE: The following prevents the default from being updated by subsequent assignment
        #                     (in this case, to [OUTCOME, {NAME= MSE}]), but fails to expose default in IDE
        # output_ports = output_ports or [OUTCOME, MSE]

        super().__init__(monitor=input_ports,
                         function=function,
                         output_ports=output_ports, # prevent default from getting overwritten by later assign
                         params=params,
                         name=name,
                         prefs=prefs,
                         **kwargs
                         )

        # Require Projection to TARGET InputPort (already required for SAMPLE as primary InputPort)
        self.input_ports[1].parameters.require_projection_in_composition.set(True, override=True)
        # reset input ports spec to actual spec, not parsed
        self.parameters.input_ports.spec = input_ports_spec

    def _validate_params(self, request_set, target_set=None, context=None):
        """If sample and target values are specified, validate that they are compatible
        """

        if INPUT_PORTS in request_set and request_set[INPUT_PORTS] is not None:
            input_ports = request_set[INPUT_PORTS]

            # Validate that there are exactly two input_ports (for sample and target)
            num_input_ports = len(input_ports)
            if num_input_ports != 2:
                raise ComparatorMechanismError(f"{INPUT_PORTS} arg is specified for {self.__class__.__name__} "
                                               f"({len(input_ports)}), so it must have exactly 2 items, "
                                               f"one each for {SAMPLE} and {TARGET}.")

            # Validate that input_ports are specified as dicts
            if not all(isinstance(input_port,dict) for input_port in input_ports):
                raise ComparatorMechanismError("PROGRAM ERROR: all items in input_port args must be converted to dicts"
                                               " by calling Port._parse_port_spec() before calling super().__init__")

            # Validate length of variable for sample = target
            if VARIABLE in input_ports[0]:
                # input_ports arg specified in standard port specification dict format
                lengths = [len(input_port[VARIABLE]) if input_port[VARIABLE] is not None else 0
                           for input_port in input_ports]
            else:
                # input_ports arg specified in {<Port_Name>:<PORT SPECIFICATION DICT>} format
                lengths = [len(list(input_port_dict.values())[0][VARIABLE]) for input_port_dict in input_ports]

            if lengths[0] != lengths[1]:
                raise ComparatorMechanismError(f"Length of value specified for {SAMPLE} InputPort "
                                               f"of {self.__class__.__name__} ({lengths[0]}) must be "
                                               f"same as length of value specified for {TARGET} ({lengths[1]}).")

        elif SAMPLE in request_set and TARGET in request_set:

            sample = request_set[SAMPLE]
            if isinstance(sample, InputPort):
                sample_value = sample.value
            elif isinstance(sample, Mechanism):
                sample_value = sample.input_value[0]
            elif is_value_spec(sample):
                sample_value = sample
            else:
                sample_value = None

            target = request_set[TARGET]
            if isinstance(target, InputPort):
                target_value = target.value
            elif isinstance(target, Mechanism):
                target_value = target.input_value[0]
            elif is_value_spec(target):
                target_value = target
            else:
                target_value = None

            if sample is not None and target is not None:
                if not iscompatible(sample, target, **{kwCompatibilityLength: True,
                                                       kwCompatibilityNumeric: True}):
                    raise ComparatorMechanismError(f"The length of the sample ({len(sample)}) "
                                                   f"must be the same as for the target ({len(target)})"
                                                   f"for {self.__class__.__name__} {self.name}.")

        super()._validate_params(request_set=request_set,
                                 target_set=target_set,
                                 context=context)

    def _merge_legacy_constructor_args(self, sample, target, default_variable=None, input_ports=None):

        # USE sample and target TO CREATE AN InputPort specfication dictionary for each;
        # DO SAME FOR InputPorts argument, USE TO OVERWRITE ANY SPECIFICATIONS IN sample AND target DICTS
        # TRY tuple format AS WAY OF PROVIDED CONSOLIDATED variable AND OutputPort specifications

        sample_dict = _parse_port_spec(owner=self,
                                        port_type=InputPort,
                                        port_spec=sample,
                                        name=SAMPLE)

        target_dict = _parse_port_spec(owner=self,
                                        port_type=InputPort,
                                        port_spec=target,
                                        name=TARGET)

        # If either the default_variable arg or the input_ports arg is provided:
        #    - validate that there are exactly two items in default_variable or input_ports list
        #    - if there is an input_ports list, parse it and use it to update sample and target dicts
        if input_ports:
            input_ports = input_ports[INPUT_PORTS]
            # print("type input_ports = {}".format(type(input_ports)))
            if not isinstance(input_ports, list):
                raise ComparatorMechanismError(f"If an '{INPUT_PORTS}' argument is included in the constructor "
                                               f"for a {ComparatorMechanism.__name__} it must be a list with "
                                               f"two {InputPort.__name__} specifications.")

        input_ports = input_ports or default_variable

        if input_ports is not None:
            if len(input_ports)!=2:
                raise ComparatorMechanismError(f"If an \'input_ports\' arg is included in the constructor for a "
                                               f"{ComparatorMechanism.__name__}, it must be a list with exactly "
                                               f"two items (not {len(input_ports)}).")

            sample_input_port_dict = _parse_port_spec(owner=self,
                                                        port_type=InputPort,
                                                        port_spec=input_ports[0],
                                                        name=SAMPLE,
                                                        value=None)

            target_input_port_dict = _parse_port_spec(owner=self,
                                                        port_type=InputPort,
                                                        port_spec=input_ports[1],
                                                        name=TARGET,
                                                        value=None)

            sample_dict = recursive_update(sample_dict, sample_input_port_dict)
            target_dict = recursive_update(target_dict, target_input_port_dict)

        return [sample_dict, target_dict]



import numpy as np
import psyneulink as pnl
import pytest


@pytest.mark.parametrize(
    'projection_type, sender_variable, receiver_variable, projection_value, function_value',
    [
        (pnl.MappingProjection, [0, 0, 0], [0, 0], np.array([0, 0]), np.array([0, 0]))
    ]
)
def test_value_shapes_with_matrix(projection_type, sender_variable, receiver_variable, projection_value, function_value):
    A = pnl.TransferMechanism(default_variable=sender_variable)
    B = pnl.TransferMechanism(default_variable=receiver_variable)
    P = projection_type(sender=A, receiver=B)

    assert P.defaults.value.shape == projection_value.shape
    assert P.function.defaults.value.shape == function_value.shape



from psyneulink.core.compositions.report import ReportOutput, ReportProgress
import psyneulink as pnl

# Settings for running script:

# MODEL_PARAMS = 'TestParams'
MODEL_PARAMS = 'DeclanParams'

CONSTRUCT_MODEL = True                 # THIS MUST BE SET TO True to run the script
DISPLAY_MODEL =  (                     # Only one of the following can be uncommented:
    None                             # suppress display of model
    # {                                  # show simple visual display of model
    #     # 'show_pytorch': True,            # show pytorch graph of model
    #     'show_learning': True,
    #     # 'show_nested_args': {'show_learning': pnl.ALL},
    #     # 'show_projections_not_in_composition': True,
    #     # 'show_nested': {'show_node_structure': True},
    #     # 'exclude_from_gradient_calc_style': 'dashed'# show target mechanisms for learning
    #     # 'show_node_structure': True     # show detailed view of node structures and projections
    # }
)
# RUN_MODEL = False                      # False => don't run the model
RUN_MODEL = True,                       # True => run the model
# REPORT_OUTPUT = ReportOutput.FULL  # Sets console output during run [ReportOutput.ON, .TERSE OR .FULL]
REPORT_OUTPUT = ReportOutput.OFF     # Sets console output during run [ReportOutput.ON, .TERSE OR .FULL]
REPORT_PROGRESS = ReportProgress.OFF # Sets console progress bar during run
PRINT_RESULTS = False                # don't print model.results to console after execution
# PRINT_RESULTS = True                 # print model.results to console after execution
SAVE_RESULTS = False                 # save model.results to disk
# PLOT_RESULTS = False                  # don't plot results (PREDICTIONS) vs. TARGETS
PLOT_RESULTS = True                  # plot results (PREDICTIONS) vs. TARGETS
ANIMATE = False                       # {UNIT:EXECUTION_SET} # Specifies whether to generate animation of execution



import numpy as np
import matplotlib.pyplot as plt
#matplotlib inline
import psyneulink as pnl
import psyneulink.core.components.functions.nonstateful.transferfunctions
from psyneulink.core.components.functions.nonstateful.learningfunctions import BackPropagation

nouns = ['oak', 'pine', 'rose', 'daisy', 'canary', 'robin', 'salmon', 'sunfish']
relations = ['is', 'has', 'can']
is_list = ['living', 'living thing', 'plant', 'animal', 'tree', 'flower', 'bird', 'fish', 'big', 'green', 'red',
           'yellfsow']
has_list = ['roots', 'leaves', 'bark', 'branches', 'skin', 'feathers', 'wings', 'gills', 'scales']
can_list = ['grow', 'move', 'swim', 'fly', 'breathe', 'breathe underwater', 'breathe air', 'walk', 'photosynthesize']
descriptors = [nouns, is_list, has_list, can_list]

truth_nouns = np.identity(len(nouns))

truth_is = np.zeros((len(nouns), len(is_list)))

truth_is[0, :] = [1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0]
truth_is[1, :] = [1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0]
truth_is[2, :] = [1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0]
truth_is[3, :] = [1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0]
truth_is[4, :] = [1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1]
truth_is[5, :] = [1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1]
truth_is[6, :] = [1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 1, 0]
truth_is[7, :] = [1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0]

truth_has = np.zeros((len(nouns), len(has_list)))

truth_has[0, :] = [1, 1, 1, 1, 0, 0, 0, 0, 0]
truth_has[1, :] = [1, 1, 1, 1, 0, 0, 0, 0, 0]
truth_has[2, :] = [1, 1, 0, 0, 0, 0, 0, 0, 0]
truth_has[3, :] = [1, 1, 0, 0, 0, 0, 0, 0, 0]
truth_has[4, :] = [0, 0, 0, 0, 1, 1, 1, 0, 0]
truth_has[5, :] = [0, 0, 0, 0, 1, 1, 1, 0, 0]
truth_has[6, :] = [0, 0, 0, 0, 0, 0, 0, 1, 1]
truth_has[7, :] = [0, 0, 0, 0, 0, 0, 0, 1, 1]

truth_can = np.zeros((len(nouns), len(can_list)))

truth_can[0, :] = [1, 0, 0, 0, 0, 0, 0, 0, 1]
truth_can[1, :] = [1, 0, 0, 0, 0, 0, 0, 0, 1]
truth_can[2, :] = [1, 0, 0, 0, 0, 0, 0, 0, 1]
truth_can[3, :] = [1, 0, 0, 0, 0, 0, 0, 0, 1]
truth_can[4, :] = [1, 1, 0, 1, 1, 0, 1, 1, 0]
truth_can[5, :] = [1, 1, 0, 1, 1, 0, 1, 1, 0]
truth_can[6, :] = [1, 1, 1, 0, 1, 1, 0, 0, 0]
truth_can[7, :] = [1, 1, 1, 0, 1, 1, 0, 0, 0]

truths = [[truth_nouns], [truth_is], [truth_has], [truth_can]]

def gen_input_vals(nouns, relations):

    X_1=np.vstack((np.identity(len(nouns)),np.ones((1,len(nouns)))))
    X_1=X_1.T

    X_2=np.vstack((np.identity(len(relations)),np.ones((1,len(relations)))))
    X_2=X_2.T
    return (X_1, X_2)

nouns_onehot, rels_onehot = gen_input_vals(nouns, relations)

r_1 = np.shape(nouns_onehot)[0]
c_1 = np.shape(nouns_onehot)[1]
r_2 = np.shape(rels_onehot)[0]
c_2 = np.shape(rels_onehot)[1]

############### THIS IS THE PART WHERE WE START BUILDING TRANSFER MECHANISMS ########################

#In order to build in biases, add an extra node to every layer, including the inputs

nouns_in = pnl.TransferMechanism(name="nouns_input",
                                 default_variable=np.zeros(c_1)
                                )

rels_in = pnl.TransferMechanism(name="rels_input",
                                default_variable=np.zeros(c_2)
                               )

h1 = pnl.TransferMechanism(name="hidden_nouns",
                           input_shapes=9,
                           function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic()
                            )

h2 = pnl.TransferMechanism(name="hidden_mixed",
                           input_shapes=16,
                           function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic()
                               )

out_sig_I = pnl.TransferMechanism(name="sig_outs_I",
                                  input_shapes=len(nouns),
                                  function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic()
                                  )

out_sig_is = pnl.TransferMechanism(name="sig_outs_is",
                                   input_shapes=len(is_list),
                                   function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic()
 )

out_sig_has = pnl.TransferMechanism(name="sig_outs_has",
                                    input_shapes=len(has_list),
                                    function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic()
                                    )

out_sig_can = pnl.TransferMechanism(name="sig_outs_can",
                                    input_shapes=len(can_list),
                                    function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic()
                                    )

###################### THIS IS THE PART WHERE I PUT IN THE FORCED RANDOM MATRICES #########################

#alla de maps

map_nouns_h1 = pnl.MappingProjection(matrix=np.random.rand(c_1,c_1),
                                name="map_nouns_h1"
                                )

map_rel_h2 = pnl.MappingProjection(matrix=np.random.rand(c_2,16),
                                name="map_relh2"
                                )

map_h1_h2 = pnl.MappingProjection(matrix=np.random.rand(c_1,16),
                                name="map_h1_h2"
                                )

map_h2_I = pnl.MappingProjection(matrix=np.random.rand(16,len(nouns)),
                                name="map_h2_I"
                                )

map_h2_is = pnl.MappingProjection(matrix=np.random.rand(16,len(is_list)),
                                name="map_h2_is"
                                )

map_h2_has = pnl.MappingProjection(matrix=np.random.rand(16,len(has_list)),
                                name="map_h2_has"
                                )

map_h2_can = pnl.MappingProjection(matrix=np.random.rand(16,len(can_list)),
                                name="map_h2_can"
                                )

#################### THIS IS THE PART WHERE WE START BUILDING OUT ALL THE PROCESSES ########################

p11 = pnl.Pathway(pathway=[nouns_in,
                           map_nouns_h1,
                           h1,
                           map_h1_h2,
                           h2])

p12 = pnl.Pathway(pathway=[rels_in,
                           map_rel_h2,
                           h2])

p21 = pnl.Pathway(pathway=[h2,
                           map_h2_I,
                           out_sig_I])

p22 = pnl.Pathway(pathway=[h2,
                           map_h2_is,
                           out_sig_is])

p23 = pnl.Pathway(pathway=[h2,
                           map_h2_has,
                           out_sig_has])

p24 = pnl.Pathway(pathway=[h2,
                           map_h2_can,
                           out_sig_can])

############################# THIS IS WHERE WE BUILD OUT THE COMPOSITION ###################################

rumel_comp = pnl.Composition(pathways=[(p11, BackPropagation),
                                       (p12, BackPropagation),
                                       (p21, BackPropagation),
                                       (p22, BackPropagation),
                                       (p23, BackPropagation),
                                       (p24, BackPropagation),
                                       ],
                             learning_rate=.5)

rumel_comp.show_graph(output_fmt='jupyter')

############################## THIS IS WHERE WE SETUP THE LOOP VARIABLES #########################################

# THESE ARRAYS STORE THE ERROR VALUES FROM THE SIG OUTPUT AND BIN OUTPUT
delta_bin_array=[]
delta_sig_array=[]

# SET NUMBER OF EPOCHS:
epochs=1000

# CREATE CONDITIONAL:
div = epochs / 100
spits=np.arange(0,epochs,div)

#CREATE KILLSWITCH:
kill=0

############################## THIS IS WHERE WE RUN THE COMPOSITION #########################################

for epoch in range(epochs):
    print("epoch number", epoch)
    for noun in range(len(nouns)):
        for rel_out in range(3):
            # K GIVES THE OUTPUT OF THE COMPOSITION

            k = rumel_comp.learn(inputs={nouns_in: nouns_onehot[noun],
                                         rels_in: rels_onehot[rel_out],
                                         },
                                 targets={out_sig_I: truth_nouns[noun],
                                          out_sig_is: truth_is[noun],
                                          out_sig_has: truth_has[noun],
                                          out_sig_can: truth_can[noun]
                                          },
                                 )

            # PUT K INTO AN ARRAY SO WE CAN MANIPULATE ITS VALUES

            k_array = np.array(k)

            # IT_K GIVES THE OUTPUT FROM THIS SPECIFIC RUN

            it_k = k[np.shape(k_array)[0] - 1]

            # THE DELTAS ADD UP THE SQUARED ERROR FROM EVERY OUTPUT OF K (I, IS, HAS, AND CAN)

            delta = 0
            delta = np.sum((truth_nouns[noun] - np.round(it_k[0])) ** 2)
            delta = delta + np.sum((truth_is[noun] - np.round(it_k[1])) ** 2)
            delta = delta + np.sum((truth_has[noun] - np.round(it_k[2])) ** 2)
            delta = delta + np.sum((truth_can[noun] - np.round(it_k[3])) ** 2)
            delta = delta / (len(nouns) + len(is_list) + len(has_list) + len(can_list))

            delta_sig = 0
            delta_sig = np.sum((truth_nouns[noun] - (it_k[0])) ** 2)
            delta_sig = delta_sig + np.sum((truth_is[noun] - (it_k[1])) ** 2)
            delta_sig = delta_sig + np.sum((truth_has[noun] - (it_k[2])) ** 2)
            delta_sig = delta_sig + np.sum((truth_can[noun] - (it_k[3])) ** 2)
            delta_sig = delta_sig / (len(nouns) + len(is_list) + len(has_list) + len(can_list))

            # THE ARRAYS STORE THE ERROR FROM EVERY RUN. TO SMOOTH THESE, WE CAN AVERAGE THEM OVER EPOCHS.

            delta_bin_array = np.append(delta_bin_array, delta)
            delta_sig_array = np.append(delta_sig_array, delta_sig)

    # PRINT PROGRESS INFORMATION

    if np.isin(epoch, spits):
        print('the average sum squared error on sigmoids for this epoch was', np.sum(delta_sig_array[-25:]) / 24)
        print('the average sum squared error on binaries for this epoch was', np.sum(delta_bin_array[-25:]) / 24)

        # KILL THE LOOP ONCE THE LABELS CONVERGE TO ZERO ERROR FOR A CERTAIN NUMBER OF EPOCHS

    if (np.sum(delta_bin_array[-25:]) / 24) == 0.0:
        kill = kill + 1
        if kill >= 99:
            break

######################## SETUP THE LABEL ERRORS TO BE GRAPHED #######################################

delta_bin_array=np.array(delta_bin_array)
delta_bin_array_trunc = np.array(delta_bin_array[0:int(np.floor(len(delta_bin_array) / 24) * 24)])
height = len(delta_bin_array_trunc) / 24
delta_bin_array_trunc=np.reshape(delta_bin_array_trunc,(int(height),24))
delta_bin_epochs = np.sum(delta_bin_array_trunc, axis=1) / 24

######################## SETUP THE SIGMOID ERRORS TO BE GRAPHED #######################################

delta_sig_array=np.array(delta_sig_array)
delta_sig_array_trunc = delta_sig_array[0:int(np.floor(len(delta_sig_array) / 24) * 24)]
delta_sig_array_trunc=np.reshape(delta_sig_array_trunc,(int(height),24))
delta_sig_epochs = np.sum(delta_sig_array_trunc, axis=1) / 24

######################## DO THE PLOTTING #######################################

plt.plot(delta_bin_epochs)
plt.title('Label error as a function of epochs')
plt.ylabel('Error')
plt.xlabel('Epoch')
plt.show()

plt.plot(delta_sig_epochs)
plt.title('sigmoid error as a function of epochs')
plt.ylabel('Error')
plt.xlabel('Epoch')
plt.show()



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* LLVM bindings **************************************************************

import atexit
import ctypes
import enum
import functools
import inspect
from llvmlite import ir
import numpy as np
import os
import re
import time
import weakref

from psyneulink._typing import Set
from psyneulink.core.scheduling.time import Time, TimeScale
from psyneulink.core.globals.sampleiterator import SampleIterator
from psyneulink.core.globals.utilities import ContentAddressableList
from psyneulink.core import llvm as pnlvm

from . import codegen
from . import helpers
from .debug import debug_env

__all__ = ['LLVMBuilderContext', '_modules', '_find_llvm_function']


_modules: Set[ir.Module] = set()
_all_modules: Set[ir.Module] = set()
_struct_count = 0


@atexit.register
def module_count():
    if "stat" in debug_env:
        print("Active LLVM modules: ", len(_all_modules))
        print("Total ctype structures created: ", _struct_count)


_BUILTIN_PREFIX = "__pnl_builtin_"
_builtin_intrinsics = frozenset(('pow', 'log', 'exp', 'tanh', 'coth', 'csch', 'sin', 'cos',
                                 'is_close_float', 'is_close_double',
                                 'mt_rand_init', 'philox_rand_init',
                                 'get_printf_address'))


class _node_assembly():
    def __init__(self, composition, node):
        self._comp = weakref.proxy(composition)
        self._node = node

    def __repr__(self):
        return "Node wrapper for node '{}' in composition '{}'".format(self._node, self._comp)

    def _gen_llvm_function(self, *, ctx, tags:frozenset):
        return codegen.gen_node_assembly(ctx, self._comp, self._node, tags=tags)

def _comp_cached(func):
    @functools.wraps(func)
    def wrapper(bctx, obj):
        bctx._stats[func.__name__ + "_requests"] += 1
        try:
            obj_cache = bctx._cache.setdefault(obj, dict())
        except TypeError:  # 'super()' references can't be cached
            obj_cache = None
        else:
            if func in obj_cache:
                return obj_cache[func]

        bctx._stats[func.__name__ + "_misses"] += 1
        val = func(bctx, obj)
        if obj_cache is not None:
            obj_cache[func] = val
        return val

    return wrapper


class LLVMBuilderContext:
    __current_context = None
    __uniq_counter = 0
    _llvm_generation = 0
    int32_ty = ir.IntType(32)
    default_float_ty = ir.DoubleType()
    bool_ty = ir.IntType(1)

    def __init__(self, float_ty):
        assert LLVMBuilderContext.__current_context is None
        self._modules = []
        self._cache = weakref.WeakKeyDictionary()
        self._component_param_use = weakref.WeakKeyDictionary()
        self._component_state_use = weakref.WeakKeyDictionary()

        # Supported stats are listed explicitly to catch typos
        self._stats = { "function_cache_misses":0,
                        "function_cache_requests":0,
                        "types_converted":0,
                        "get_param_struct_type_misses":0,
                        "get_state_struct_type_misses":0,
                        "get_data_struct_type_misses":0,
                        "get_input_struct_type_misses":0,
                        "get_output_struct_type_misses":0,
                        "get_param_struct_type_requests":0,
                        "get_state_struct_type_requests":0,
                        "get_data_struct_type_requests":0,
                        "get_input_struct_type_requests":0,
                        "get_output_struct_type_requests":0,
                      }
        self.float_ty = float_ty
        self.init_builtins()
        LLVMBuilderContext.__current_context = self

    def __enter__(self):
        module = ir.Module(name="PsyNeuLinkModule-" + str(LLVMBuilderContext._llvm_generation))
        self._modules.append(module)
        LLVMBuilderContext._llvm_generation += 1
        return self

    def __exit__(self, e_type, e_value, e_traceback):
        assert len(self._modules) > 0
        module = self._modules.pop()
        _modules.add(module)
        _all_modules.add(module)

    @property
    def module(self):
        assert len(self._modules) > 0
        return self._modules[-1]

    def print_stats(self):
        def _hit_rate(reqs, misses):
            return round((reqs - misses) / reqs * 100, 2)

        print("LLVM codegen stats for context:", hex(id(self)))
        print("  Last compiled generation: {}".format(self._llvm_generation))

        req_stat = self._stats["function_cache_requests"]
        miss_stat = self._stats["function_cache_misses"]
        print("  Object function cache: {} requests, {} misses (hr: {}%)".format(req_stat, miss_stat, _hit_rate(req_stat, miss_stat)))
        for stat in ("input", "output", "param", "state", "data"):
            req_stat = self._stats["get_{}_struct_type_requests".format(stat)]
            miss_stat = self._stats["get_{}_struct_type_misses".format(stat)]
            print("  Total {} struct types requested from global context: {}, generated: {} (hr: {}%)".format(
                  stat, req_stat, miss_stat, _hit_rate(req_stat, miss_stat)))
        print("  Total python types converted by global context: {}".format(self._stats["types_converted"]))


    def __del__(self):
        if "stat" in debug_env:
            self.print_stats()

    @classmethod
    def get_current(cls):
        if cls.__current_context is None:
            return LLVMBuilderContext(cls.default_float_ty)
        return cls.__current_context

    @classmethod
    def is_active(cls):
        return cls.__current_context is not None

    @classmethod
    def clear_global(cls):
        cls.__current_context = None

    @classmethod
    def get_unique_name(cls, name: str):
        cls.__uniq_counter += 1
        name = re.sub(r"[^a-zA-Z0-9_]", "_", name)
        return name + '_' + str(cls.__uniq_counter)

    def init_builtins(self):
        start = time.perf_counter()
        with self as ctx:
            # Numeric
            pnlvm.builtins.setup_pnl_intrinsics(ctx)
            pnlvm.builtins.setup_csch(ctx)
            pnlvm.builtins.setup_coth(ctx)
            pnlvm.builtins.setup_tanh(ctx)
            pnlvm.builtins.setup_is_close(ctx)

            # PRNG
            pnlvm.builtins.setup_mersenne_twister(ctx)
            pnlvm.builtins.setup_philox(ctx)

            # Matrix/Vector
            pnlvm.builtins.setup_vxm(ctx)
            pnlvm.builtins.setup_vxm_transposed(ctx)
            pnlvm.builtins.setup_vec_add(ctx)
            pnlvm.builtins.setup_vec_sum(ctx)
            pnlvm.builtins.setup_mat_add(ctx)
            pnlvm.builtins.setup_vec_sub(ctx)
            pnlvm.builtins.setup_mat_sub(ctx)
            pnlvm.builtins.setup_vec_hadamard(ctx)
            pnlvm.builtins.setup_mat_hadamard(ctx)
            pnlvm.builtins.setup_vec_scalar_mult(ctx)
            pnlvm.builtins.setup_mat_scalar_mult(ctx)
            pnlvm.builtins.setup_mat_scalar_add(ctx)

        finish = time.perf_counter()

        if "time_stat" in debug_env:
            print("Time to setup PNL builtins: {}".format(finish - start))

    def get_rand_int_function_by_state(self, state):
        if len(state.type.pointee) == 5:
            return self.import_llvm_function("__pnl_builtin_mt_rand_int32_bounded")

        elif len(state.type.pointee) == 7:
            # we have different versions based on selected FP precision
            return self.import_llvm_function("__pnl_builtin_philox_rand_int32_bounded")

        else:
            assert False, "Unknown PRNG type!"

    def get_uniform_dist_function_by_state(self, state):
        if len(state.type.pointee) == 5:
            return self.import_llvm_function("__pnl_builtin_mt_rand_double")

        elif len(state.type.pointee) == 7:
            # we have different versions based on selected FP precision
            return self.import_llvm_function("__pnl_builtin_philox_rand_{}".format(str(self.float_ty)))

        else:
            assert False, "Unknown PRNG type!"

    def get_binomial_dist_function_by_state(self, state):
        if len(state.type.pointee) == 5:
            return self.import_llvm_function("__pnl_builtin_mt_rand_binomial")

        elif len(state.type.pointee) == 7:
            return self.import_llvm_function("__pnl_builtin_philox_rand_binomial")

        else:
            assert False, "Unknown PRNG type!"

    def get_normal_dist_function_by_state(self, state):
        if len(state.type.pointee) == 5:
            return self.import_llvm_function("__pnl_builtin_mt_rand_normal")

        elif len(state.type.pointee) == 7:
            # Normal exists only for self.float_ty
            return self.import_llvm_function("__pnl_builtin_philox_rand_normal")

        else:
            assert False, "Unknown PRNG type!"

    def get_builtin(self, name: str, args=[], function_type=None):
        if name in _builtin_intrinsics:
            return self.import_llvm_function(_BUILTIN_PREFIX + name)
        if name in ('maxnum'):
            function_type = ir.FunctionType(args[0], [args[0], args[0]])
        return self.module.declare_intrinsic("llvm." + name, args, function_type)

    def create_llvm_function(self, args, component, name=None, *, return_type=ir.VoidType(), tags:frozenset=frozenset()):
        name = "_".join((str(component), *tags)) if name is None else name

        # Builtins are already unique and need to keep their special name
        func_name = name if name.startswith(_BUILTIN_PREFIX) else self.get_unique_name(name)
        func_ty = ir.FunctionType(return_type, args)
        llvm_func = ir.Function(self.module, func_ty, name=func_name)
        llvm_func.attributes.add('argmemonly')
        for a in llvm_func.args:
            if isinstance(a.type, ir.PointerType):
                a.attributes.add('nonnull')

        metadata = self.get_debug_location(llvm_func, component)
        scope = dict(metadata.operands)["scope"]
        llvm_func.set_metadata("dbg", scope)

        # Create entry block
        block = llvm_func.append_basic_block(name="entry")
        builder = ir.IRBuilder(block)
        builder.debug_metadata = metadata

        return builder

    def gen_llvm_function(self, obj, *, tags:frozenset) -> ir.Function:
        obj_cache = self._cache.setdefault(obj, dict())

        self._stats["function_cache_requests"] += 1
        if tags not in obj_cache:
            self._stats["function_cache_misses"] += 1
            with self:
                obj_cache[tags] = obj._gen_llvm_function(ctx=self, tags=tags)
                self.check_used_params(obj, tags=tags)

        return obj_cache[tags]

    def import_llvm_function(self, fun, *, tags:frozenset=frozenset()) -> ir.Function:
        """
        Get function handle if function exists in current modele.
        Create function declaration if it exists in a older module.
        """
        if isinstance(fun, str):
            f = _find_llvm_function(fun, _all_modules | {self.module})
        else:
            f = self.gen_llvm_function(fun, tags=tags)

        # Add declaration to the current module
        if f.name not in self.module.globals:
            decl_f = ir.Function(self.module, f.type.pointee, f.name)
            assert decl_f.is_declaration
            return decl_f
        return f

    def get_random_state_ptr(self, builder, component, state, params):
        random_state_ptr = self.get_param_or_state_ptr(builder, component, "random_state", state_struct_ptr=state)

        # Used seed is the last member of both MT state and Philox state
        seed_idx = len(random_state_ptr.type.pointee) - 1
        used_seed_ptr = builder.gep(random_state_ptr, [self.int32_ty(0), self.int32_ty(seed_idx)])
        used_seed = builder.load(used_seed_ptr)

        seed_ptr = self.get_param_or_state_ptr(builder, component, "seed", param_struct_ptr=params)
        new_seed = pnlvm.helpers.load_extract_scalar_array_one(builder, seed_ptr)
        # FIXME: The seed should ideally be integer already.
        #        However, it can be modulated and we don't support
        #        passing integer values as computed results.
        new_seed = builder.fptoui(new_seed, used_seed.type)

        seeds_cmp = builder.icmp_unsigned("!=", used_seed, new_seed)
        with builder.if_then(seeds_cmp, likely=False):
            if seed_idx == 4:
                reseed_f = self.get_builtin("mt_rand_init")
            elif seed_idx == 6:
                reseed_f = self.get_builtin("philox_rand_init")
            else:
                assert False, "Unknown PRNG type!"

            builder.call(reseed_f, [random_state_ptr, new_seed])

        return random_state_ptr

    def get_param_or_state_ptr(self, builder, component, param, *, param_struct_ptr=None, state_struct_ptr=None, history=0):
        param_name = getattr(param, "name", param)
        param = None
        state = None

        if param_name in component.llvm_param_ids:
            assert param_struct_ptr is not None, "Can't get param ptr for: {}".format(param_name)
            self._component_param_use.setdefault(component, set()).add(param_name)
            param = helpers.get_param_ptr(builder, component, param_struct_ptr, param_name)

        if param_name in component.llvm_state_ids:
            assert state_struct_ptr is not None, "Can't get state ptr for: {}".format(param_name)
            self._component_state_use.setdefault(component, set()).add(param_name)
            state = helpers.get_state_ptr(builder, component, state_struct_ptr, param_name, history)

        if param is not None and state is not None:
            return (param, state)

        return param or state

    def get_state_space(self, builder, component, state_ptr, param):
        param_name = getattr(param, "name", param)
        self._component_state_use.setdefault(component, set()).add(param_name)
        return helpers.get_state_space(builder, component, state_ptr, param_name)

    def check_used_params(self, component, *, tags:frozenset):
        """
        This function checks that parameters included in the compiled structures are used in compiled code.

        If the assertion in this function triggers the parameter name should be added to the parameter
        block list in the Component class.
        """

        # Skip the check if the parameter use is not tracked. Some components (like node wrappers)
        # don't even have parameters.
        if component not in self._component_state_use and component not in self._component_param_use:
            return

        # Skip the check for variant functions
        if len(tags) != 0:
            return

        component_param_ids = set(component.llvm_param_ids)
        component_state_ids = set(component.llvm_state_ids)

        used_param_ids = self._component_param_use.get(component, set())
        used_state_ids = self._component_state_use.get(component, set())

        # initializers are  only used in "reset" variants
        initializers = {p.initializer for p in component.parameters}

        # has_initializers is only used in "reset" variants
        initializers.add('has_initializers')

        # 'termination_measure" is only used in "is_finished" variant
        used_param_ids.add('termination_measure')
        used_state_ids.add('termination_measure')

        # 'num_trials_per_estimate' is only used in "evaluate" variants
        if hasattr(component, 'evaluate_agent_rep'):
            used_param_ids.add('num_trials_per_estimate')

        unused_param_ids = component_param_ids - used_param_ids - initializers
        unused_state_ids = component_state_ids - used_state_ids

        assert len(unused_param_ids) == 0 and len(unused_state_ids) == 0, \
            "Compiled component '{}'(tags: {}) unused parameters: {}, state: {}".format(component, list(tags), unused_param_ids, unused_state_ids)

    @staticmethod
    def get_debug_location(func: ir.Function, component):
        mod = func.module
        path = inspect.getfile(component.__class__) if component is not None else "<pnl_builtin>"
        d_version = mod.add_metadata([ir.IntType(32)(2), "Dwarf Version", ir.IntType(32)(4)])
        di_version = mod.add_metadata([ir.IntType(32)(2), "Debug Info Version", ir.IntType(32)(3)])
        flags = mod.add_named_metadata("llvm.module.flags")
        if len(flags.operands) == 0:
            flags.add(d_version)
            flags.add(di_version)
        cu = mod.add_named_metadata("llvm.dbg.cu")
        di_file = mod.add_debug_info("DIFile", {
            "filename": os.path.basename(path),
            "directory": os.path.dirname(path),
        })
        di_func_type = mod.add_debug_info("DISubroutineType", {
            # None as `null`
            "types": mod.add_metadata([None]),
        })
        di_compileunit = mod.add_debug_info("DICompileUnit", {
            "language": ir.DIToken("DW_LANG_Python"),
            "file": di_file,
            "producer": "PsyNeuLink",
            "runtimeVersion": 0,
            "isOptimized": False,
        }, is_distinct=True)
        cu.add(di_compileunit)
        di_func = mod.add_debug_info("DISubprogram", {
            "name": func.name,
            "file": di_file,
            "line": 0,
            "type": di_func_type,
            "isLocal": False,
            "unit": di_compileunit,
        }, is_distinct=True)
        di_loc = mod.add_debug_info("DILocation", {
            "line": 0, "column": 0, "scope": di_func,
        })
        return di_loc

    @staticmethod
    def update_debug_loc_position(di_loc: ir.DIValue, line:int, column:int):
        di_func = dict(di_loc.operands)["scope"]

        return di_loc.parent.add_debug_info("DILocation", {
            "line": line, "column": column, "scope": di_func,
        })

    @_comp_cached
    def get_input_struct_type(self, component):
        if hasattr(component, '_get_input_struct_type'):
            return component._get_input_struct_type(self)

        default_var = component.defaults.variable
        return self.convert_python_struct_to_llvm_ir(default_var)

    @_comp_cached
    def get_output_struct_type(self, component):
        if hasattr(component, '_get_output_struct_type'):
            return component._get_output_struct_type(self)

        default_val = component.defaults.value
        return self.convert_python_struct_to_llvm_ir(default_val)

    @_comp_cached
    def get_param_struct_type(self, component):
        if hasattr(component, '_get_param_struct_type'):
            return component._get_param_struct_type(self)

        def _param_struct(p):
            val = p.get(None)   # this should use defaults
            if hasattr(val, "_get_compilation_params") or \
               hasattr(val, "_get_param_struct_type"):
                return self.get_param_struct_type(val)
            if isinstance(val, ContentAddressableList):
                return ir.LiteralStructType(self.get_param_struct_type(x) for x in val)
            elif p.name == 'matrix':   # Flatten matrix
                val = np.asarray(val, dtype=float).ravel()
            elif p.name == 'num_trials_per_estimate':  # Should always be int
                val = np.int32(0) if val is None else np.int32(val)
            elif np.ndim(val) == 0 and component._is_param_modulated(p):
                val = [val]   # modulation adds array wrap
            return self.convert_python_struct_to_llvm_ir(val)

        elements = map(_param_struct, component._get_compilation_params())
        return ir.LiteralStructType(elements)

    @_comp_cached
    def get_state_struct_type(self, component):
        if hasattr(component, '_get_state_struct_type'):
            return component._get_state_struct_type(self)

        def _state_struct(p):
            val = p.get(None)   # this should use defaults
            if hasattr(val, "_get_compilation_state") or \
               hasattr(val, "_get_state_struct_type"):
                return self.get_state_struct_type(val)
            if isinstance(val, ContentAddressableList):
                return ir.LiteralStructType(self.get_state_struct_type(x) for x in val)
            if p.name == 'matrix':   # Flatten matrix
                val = np.asarray(val, dtype=float).ravel()
            struct = self.convert_python_struct_to_llvm_ir(val)
            return ir.ArrayType(struct, p.history_min_length + 1)

        elements = map(_state_struct, component._get_compilation_state())
        return ir.LiteralStructType(elements)

    @_comp_cached
    def get_data_struct_type(self, component):
        if hasattr(component, '_get_data_struct_type'):
            return component._get_data_struct_type(self)

        return ir.LiteralStructType([])

    def get_node_assembly(self, composition, node):
        cache = getattr(composition, '_node_assemblies', None)
        if cache is None:
            cache = weakref.WeakKeyDictionary()
            setattr(composition, '_node_assemblies', cache)
        return cache.setdefault(node, _node_assembly(composition, node))

    def convert_python_struct_to_llvm_ir(self, t):
        self._stats["types_converted"] += 1
        if t is None:
            return ir.LiteralStructType([])

        elif isinstance(t, (list, tuple)):
            elems_t = [self.convert_python_struct_to_llvm_ir(x) for x in t]
            if len(elems_t) > 0 and all(x == elems_t[0] for x in elems_t):
                return ir.ArrayType(elems_t[0], len(elems_t))

            return ir.LiteralStructType(elems_t)

        elif isinstance(t, enum.Enum):
            # FIXME: Consider enums of non-int type
            assert all(round(x.value) == x.value for x in type(t))
            return self.int32_ty

        elif isinstance(t, (int, float, np.floating)):
            return self.float_ty

        elif isinstance(t, np.integer):
            # Python 'int' is handled above as it is the default type for '0'
            return ir.IntType(t.nbytes * 8)

        elif isinstance(t, np.ndarray):
            # 0d uint32 values were likely created from enums (above) and are
            # observed here after compilation sync.
            # Avoid silent promotion to float (via Python's builtin int-type)
            if t.ndim == 0 and t.dtype == np.uint32:
                return self.convert_python_struct_to_llvm_ir(t.reshape(1)[0])
            return self.convert_python_struct_to_llvm_ir(t.tolist())

        elif isinstance(t, np.random.RandomState):
            return pnlvm.builtins.get_mersenne_twister_state_struct(self)

        elif isinstance(t, np.random.Generator):
            assert isinstance(t.bit_generator, np.random.Philox)
            return pnlvm.builtins.get_philox_state_struct(self)

        elif isinstance(t, Time):
            return ir.ArrayType(self.int32_ty, len(TimeScale))

        elif isinstance(t, SampleIterator):
            if isinstance(t.generator, list):
                return ir.ArrayType(self.float_ty, len(t.generator))

            # Generic iterator is {start, increment, count}
            return ir.LiteralStructType((self.float_ty, self.float_ty, self.int32_ty))

        assert False, "Don't know how to convert {}".format(type(t))


def _find_llvm_function(name: str, mods=_all_modules) -> ir.Function:
    f = None
    for m in mods:
        if name in m.globals:
            f = m.get_global(name)

    if not isinstance(f, ir.Function):
        raise ValueError("No such function: {}".format(name))
    return f


def _gen_cuda_kernel_wrapper_module(function):
    module = ir.Module(name="wrapper_" + function.name)

    decl_f = ir.Function(module, function.type.pointee, function.name)
    assert decl_f.is_declaration
    orig_args = function.type.pointee.args

    # remove indices if this is grid_evaluate_ranged
    is_grid_ranged = len(orig_args) == 8 and isinstance(orig_args[2], ir.IntType)
    if is_grid_ranged:
        orig_args = orig_args[:2] + orig_args[4:]

    wrapper_type = ir.FunctionType(ir.VoidType(), [*orig_args, ir.IntType(32)])
    kernel_func = ir.Function(module, wrapper_type, function.name + "_cuda_kernel")
    # Add kernel mark metadata
    module.add_named_metadata("nvvm.annotations", [kernel_func, "kernel", ir.IntType(32)(1)])

    # Start the function
    block = kernel_func.append_basic_block(name="entry")
    builder = ir.IRBuilder(block)

    # Calculate global id of a thread in x dimension
    intrin_ty = ir.FunctionType(ir.IntType(32), [])
    tid_x_f = ir.Function(module, intrin_ty, "llvm.nvvm.read.ptx.sreg.tid.x")
    ntid_x_f = ir.Function(module, intrin_ty, "llvm.nvvm.read.ptx.sreg.ntid.x")
    ctaid_x_f = ir.Function(module, intrin_ty, "llvm.nvvm.read.ptx.sreg.ctaid.x")

    # Number of threads per block
    ntid = builder.call(ntid_x_f, [])
    # Thread ID in block
    tid = builder.call(tid_x_f, [])

    global_id = builder.mul(builder.call(ctaid_x_f, []), ntid)
    global_id = builder.add(global_id, tid)

    # Index all pointer arguments. Ignore the thread count argument
    args = list(kernel_func.args)[:-1]
    indexed_args = []

    # pointer args do not alias
    for a in args:
        if isinstance(a.type, ir.PointerType):
            a.attributes.add('noalias')

    def _upload_to_shared(b, ptr, length, name):
        shared = ir.GlobalVariable(module, ptr.type.pointee,
                                   name=function.name + "_shared_" + name,
                                   addrspace=3)
        shared.alignment = 128
        shared.linkage = "internal"
        shared_ptr = b.addrspacecast(shared, shared.type.pointee.as_pointer())

        char_ptr_ty = ir.IntType(8).as_pointer()
        bool_ty = ir.IntType(1)

        ptr_src = b.bitcast(ptr, char_ptr_ty)
        ptr_dst = b.bitcast(shared_ptr, char_ptr_ty)

        # the params for objectsize are:
        # * obj pointer,
        # * 0 on unknown size instead of -1,
        # * NULL ptr is unknown size
        # * evaluate size at runtime
        obj_size_ty = ir.FunctionType(ir.IntType(32), [char_ptr_ty, bool_ty, bool_ty, bool_ty])
        obj_size_f = module.declare_intrinsic("llvm.objectsize.i32", [], obj_size_ty)
        obj_size = b.call(obj_size_f, [ptr_dst, bool_ty(1), bool_ty(0), bool_ty(0)])
        obj_size = b.mul(obj_size, length)

        if "unaligned_copy" not in debug_env:
            copy_ty = ir.IntType(32)
            copy_ptr_ty = copy_ty.as_pointer()
            copy_bytes = copy_ty.width // 8
            obj_size = builder.add(obj_size, obj_size.type(copy_bytes - 1))
            obj_size = builder.udiv(obj_size, obj_size.type(copy_bytes))
            ptr_src = builder.bitcast(ptr, copy_ptr_ty)
            ptr_dst = builder.bitcast(shared_ptr, copy_ptr_ty)


        # copy data using as many threads as available in thread group
        with helpers.for_loop(b, tid, obj_size, ntid, id="copy_" + name) as (b1, i):
            src = b1.gep(ptr_src, [i])
            dst = b1.gep(ptr_dst, [i])
            b1.store(b1.load(src), dst)

        sync_threads_ty = ir.FunctionType(ir.VoidType(), [])
        sync_threads = module.declare_intrinsic("llvm.nvvm.barrier0", [], sync_threads_ty)
        builder.call(sync_threads, [])

        return b, shared_ptr

    if is_grid_ranged and "cuda_no_shared" not in debug_env:
        one = ir.IntType(32)(1)

        # Upload static RO structures
        builder, args[0] = _upload_to_shared(builder, args[0], one, "params")
        builder, args[1] = _upload_to_shared(builder, args[1], one, "state")
        builder, args[4] = _upload_to_shared(builder, args[4], one, "data")

        # TODO: Investigate benefit of uplaoding dynamic RO structures to
        #       shared memory (e.g. inputs)

    # Check global id and exit if we're over
    should_quit = builder.icmp_unsigned(">=", global_id, kernel_func.args[-1])
    with builder.if_then(should_quit):
        builder.ret_void()

    # If we're calling ranged search there are no offsets
    if is_grid_ranged:
        next_id = builder.add(global_id, global_id.type(1))
        call_args = args[:2] + [global_id, next_id] + args[2:]
        builder.call(decl_f, call_args)
        builder.ret_void()
        return module

    # Runs need special handling. data_in and data_out are one dimensional,
    # but hold entries for all parallel invocations.
    # comp_state, comp_params, comp_data, comp_in, comp_out, #trials, #inputs
    is_comp_run = len(args) == 7
    if is_comp_run:
        runs_count = builder.load(args[5])
        input_count = builder.load(args[6])

    for i, arg in enumerate(args):
        if isinstance(arg.type, ir.PointerType):
            offset = global_id
            if is_comp_run:
                # #inputs needs to be the same for comp run
                if i == 6:
                    offset = ir.IntType(32)(0)
                # data arrays need special handling
                elif i == 4:  # data_out
                    offset = builder.mul(global_id, runs_count)
                elif i == 3:  # data_in
                    offset = builder.mul(global_id, input_count)

            arg = builder.gep(arg, [offset])

        indexed_args.append(arg)

    builder.call(decl_f, indexed_args)
    builder.ret_void()

    return module


@functools.lru_cache(maxsize=128)
def _convert_llvm_ir_to_ctype(t: ir.Type):
    type_t = type(t)

    if type_t is ir.VoidType:
        return None

    elif type_t is ir.IntType:
        if t.width == 1:
            return ctypes.c_bool
        elif t.width == 8:
            return ctypes.c_uint8
        elif t.width == 16:
            return ctypes.c_uint16
        elif t.width == 32:
            return ctypes.c_uint32
        elif t.width == 64:
            return ctypes.c_uint64
        else:
            assert False, "Unknown integer type: {}".format(type_t)

    elif type_t is ir.DoubleType:
        return ctypes.c_double

    elif type_t is ir.FloatType:
        return ctypes.c_float

    elif type_t is ir.HalfType:
        # There's no half type in ctypes. Use uint16 instead.
        # User will need to do the necessary casting.
        return ctypes.c_uint16

    # Typed pointers (pre LLVM-15, or LLVM-15+ with opaque pointers disabled)
    elif hasattr(t, 'pointee'):
        pointee = _convert_llvm_ir_to_ctype(t.pointee)
        ret_t = ctypes.POINTER(pointee)

    elif type_t is ir.ArrayType:
        element_type = _convert_llvm_ir_to_ctype(t.element)
        ret_t = element_type * len(t)

    elif type_t is ir.LiteralStructType:
        global _struct_count
        uniq_name = "struct_" + str(_struct_count)
        _struct_count += 1

        field_list = []
        for i, e in enumerate(t.elements):
            # llvmlite modules get _unique string only works for symbol names
            field_uniq_name = uniq_name + "field_" + str(i)
            field_list.append((field_uniq_name, _convert_llvm_ir_to_ctype(e)))

        ret_t = type(uniq_name, (ctypes.Structure,), {"__init__": ctypes.Structure.__init__})
        ret_t._fields_ = field_list
        assert len(ret_t._fields_) == len(t.elements)

    else:
        assert False, "Don't know how to convert LLVM type: {}({})".format(t, type(t))

    return ret_t

@functools.lru_cache(maxsize=16)
def _convert_llvm_ir_to_dtype(t: ir.Type):

    if isinstance(t, ir.IntType):
        if t.width == 8:
            return np.uint8().dtype

        elif t.width == 16:
            return np.uint16().dtype

        elif t.width == 32:
            return np.uint32().dtype

        elif t.width == 64:
            return np.uint64().dtype

        else:
            assert False, "Unsupported integer type: {}".format(type(t))

    elif isinstance(t, ir.DoubleType):
        return np.float64().dtype

    elif isinstance(t, ir.FloatType):
        return np.float32().dtype

    elif isinstance(t, ir.HalfType):
        return np.float16().dtype

    elif isinstance(t, ir.ArrayType):
        element_type = _convert_llvm_ir_to_dtype(t.element)

        # Create multidimensional array instead of nesting
        if element_type.subdtype is not None:
            element_type, shape = element_type.subdtype
        else:
            shape = ()

        ret_t = np.dtype((element_type, (len(t),) + shape))

    elif isinstance(t, ir.LiteralStructType):
        field_list = []
        for i, e in enumerate(t.elements):
            field_list.append(("field_" + str(i), _convert_llvm_ir_to_dtype(e)))

        ret_t = np.dtype(field_list, align=True)
    else:
        assert False, "Don't know how to convert LLVM type to dtype: {}".format(t)

    return ret_t
